@@ -18,6 +18,8 @@
 #include <math.h>
 
 #include "common.h"
+#include "colorspace.h"
+#include "hash.h"
 
 #include <libplacebo/colorspace.h>
 #include <libplacebo/tone_mapping.h>
@@ -68,6 +70,27 @@ bool pl_color_system_is_linear(enum pl_color_system sys)
     pl_unreachable();
 }
 
+const char *const pl_color_system_names[PL_COLOR_SYSTEM_COUNT] = {
+    [PL_COLOR_SYSTEM_UNKNOWN]       = "Auto (unknown)",
+    [PL_COLOR_SYSTEM_BT_601]        = "ITU-R Rec. BT.601 (SD)",
+    [PL_COLOR_SYSTEM_BT_709]        = "ITU-R Rec. BT.709 (HD)",
+    [PL_COLOR_SYSTEM_SMPTE_240M]    = "SMPTE-240M",
+    [PL_COLOR_SYSTEM_BT_2020_NC]    = "ITU-R Rec. BT.2020 (non-constant luminance)",
+    [PL_COLOR_SYSTEM_BT_2020_C]     = "ITU-R Rec. BT.2020 (constant luminance)",
+    [PL_COLOR_SYSTEM_BT_2100_PQ]    = "ITU-R Rec. BT.2100 ICtCp PQ variant",
+    [PL_COLOR_SYSTEM_BT_2100_HLG]   = "ITU-R Rec. BT.2100 ICtCp HLG variant",
+    [PL_COLOR_SYSTEM_DOLBYVISION]   = "Dolby Vision (invalid for output)",
+    [PL_COLOR_SYSTEM_YCGCO]         = "YCgCo (derived from RGB)",
+    [PL_COLOR_SYSTEM_RGB]           = "Red, Green and Blue",
+    [PL_COLOR_SYSTEM_XYZ]           = "Digital Cinema Distribution Master (XYZ)",
+};
+
+const char *pl_color_system_name(enum pl_color_system sys)
+{
+    pl_assert(sys >= 0 && sys < PL_COLOR_SYSTEM_COUNT);
+    return pl_color_system_names[sys];
+}
+
 enum pl_color_system pl_color_system_guess_ycbcr(int width, int height)
 {
     if (width >= 1280 || height > 576) {
@@ -214,6 +237,33 @@ bool pl_color_primaries_is_wide_gamut(enum pl_color_primaries prim)
     pl_unreachable();
 }
 
+const char *const pl_color_primaries_names[PL_COLOR_PRIM_COUNT] = {
+    [PL_COLOR_PRIM_UNKNOWN]     = "Auto (unknown)",
+    [PL_COLOR_PRIM_BT_601_525]  = "ITU-R Rec. BT.601 (525-line = NTSC, SMPTE-C)",
+    [PL_COLOR_PRIM_BT_601_625]  = "ITU-R Rec. BT.601 (625-line = PAL, SECAM)",
+    [PL_COLOR_PRIM_BT_709]      = "ITU-R Rec. BT.709 (HD), also sRGB",
+    [PL_COLOR_PRIM_BT_470M]     = "ITU-R Rec. BT.470 M",
+    [PL_COLOR_PRIM_EBU_3213]    = "EBU Tech. 3213-E / JEDEC P22 phosphors",
+    [PL_COLOR_PRIM_BT_2020]     = "ITU-R Rec. BT.2020 (Ultra HD)",
+    [PL_COLOR_PRIM_APPLE]       = "Apple RGB",
+    [PL_COLOR_PRIM_ADOBE]       = "Adobe RGB (1998)",
+    [PL_COLOR_PRIM_PRO_PHOTO]   = "ProPhoto RGB (ROMM)",
+    [PL_COLOR_PRIM_CIE_1931]    = "CIE 1931 RGB primaries",
+    [PL_COLOR_PRIM_DCI_P3]      = "DCI-P3 (Digital Cinema)",
+    [PL_COLOR_PRIM_DISPLAY_P3]  = "DCI-P3 (Digital Cinema) with D65 white point",
+    [PL_COLOR_PRIM_V_GAMUT]     = "Panasonic V-Gamut (VARICAM)",
+    [PL_COLOR_PRIM_S_GAMUT]     = "Sony S-Gamut",
+    [PL_COLOR_PRIM_FILM_C]      = "Traditional film primaries with Illuminant C",
+    [PL_COLOR_PRIM_ACES_AP0]    = "ACES Primaries #0",
+    [PL_COLOR_PRIM_ACES_AP1]    = "ACES Primaries #1",
+};
+
+const char *pl_color_primaries_name(enum pl_color_primaries prim)
+{
+    pl_assert(prim >= 0 && prim < PL_COLOR_PRIM_COUNT);
+    return pl_color_primaries_names[prim];
+}
+
 enum pl_color_primaries pl_color_primaries_guess(int width, int height)
 {
     // HD content
@@ -233,6 +283,32 @@ enum pl_color_primaries pl_color_primaries_guess(int width, int height)
     }
 }
 
+const char *const pl_color_transfer_names[PL_COLOR_TRC_COUNT] = {
+    [PL_COLOR_TRC_UNKNOWN]      = "Auto (unknown SDR)",
+    [PL_COLOR_TRC_BT_1886]      = "ITU-R Rec. BT.1886 (CRT emulation + OOTF)",
+    [PL_COLOR_TRC_SRGB]         = "IEC 61966-2-4 sRGB (CRT emulation)",
+    [PL_COLOR_TRC_LINEAR]       = "Linear light content",
+    [PL_COLOR_TRC_GAMMA18]      = "Pure power gamma 1.8",
+    [PL_COLOR_TRC_GAMMA20]      = "Pure power gamma 2.0",
+    [PL_COLOR_TRC_GAMMA22]      = "Pure power gamma 2.2",
+    [PL_COLOR_TRC_GAMMA24]      = "Pure power gamma 2.4",
+    [PL_COLOR_TRC_GAMMA26]      = "Pure power gamma 2.6",
+    [PL_COLOR_TRC_GAMMA28]      = "Pure power gamma 2.8",
+    [PL_COLOR_TRC_PRO_PHOTO]    = "ProPhoto RGB (ROMM)",
+    [PL_COLOR_TRC_ST428]        = "Digital Cinema Distribution Master (XYZ)",
+    [PL_COLOR_TRC_PQ]           = "ITU-R BT.2100 PQ (perceptual quantizer), aka SMPTE ST2048",
+    [PL_COLOR_TRC_HLG]          = "ITU-R BT.2100 HLG (hybrid log-gamma), aka ARIB STD-B67",
+    [PL_COLOR_TRC_V_LOG]        = "Panasonic V-Log (VARICAM)",
+    [PL_COLOR_TRC_S_LOG1]       = "Sony S-Log1",
+    [PL_COLOR_TRC_S_LOG2]       = "Sony S-Log2",
+};
+
+const char *pl_color_transfer_name(enum pl_color_transfer trc)
+{
+    pl_assert(trc >= 0 && trc < PL_COLOR_TRC_COUNT);
+    return pl_color_transfer_names[trc];
+}
+
 // HLG 75% value (scene-referred)
 #define HLG_75 3.17955
 
@@ -277,12 +353,6 @@ const struct pl_hdr_metadata pl_hdr_metadata_hdr10 ={
     .max_fall = 0, // unknown
 };
 
-static const float PQ_M1 = 2610./4096 * 1./4,
-                   PQ_M2 = 2523./4096 * 128,
-                   PQ_C1 = 3424./4096,
-                   PQ_C2 = 2413./4096 * 32,
-                   PQ_C3 = 2392./4096 * 32;
-
 float pl_hdr_rescale(enum pl_hdr_scaling from, enum pl_hdr_scaling to, float x)
 {
     if (from == to)
@@ -471,6 +541,183 @@ bool pl_color_space_is_black_scaled(const struct pl_color_space *csp)
     pl_unreachable();
 }
 
+#define MAP3(...)                                                               \
+    do {                                                                        \
+        float X;                                                                \
+        for (int _i = 0; _i < 3; _i++) {                                        \
+            X = color[_i];                                                      \
+            color[_i] = __VA_ARGS__;                                            \
+        }                                                                       \
+    } while (0)
+
+void pl_color_linearize(const struct pl_color_space *csp, float color[3])
+{
+    if (csp->transfer == PL_COLOR_TRC_LINEAR)
+        return;
+
+    float csp_min, csp_max;
+    pl_color_space_nominal_luma_ex(pl_nominal_luma_params(
+        .color      = csp,
+        .metadata   = PL_HDR_METADATA_HDR10,
+        .scaling    = PL_HDR_NORM,
+        .out_min    = &csp_min,
+        .out_max    = &csp_max,
+    ));
+
+    MAP3(fmaxf(X, 0));
+
+    switch (csp->transfer) {
+    case PL_COLOR_TRC_SRGB:
+        MAP3(X > 0.04045f ? powf((X + 0.055f) / 1.055f, 2.4f)
+                          : X / 12.92f);
+        goto scale_out;
+    case PL_COLOR_TRC_BT_1886: {
+        const float lb = powf(csp_min, 1/2.4f);
+        const float lw = powf(csp_max, 1/2.4f);
+        const float a = powf(lw - lb, 2.4f);
+        const float b = lb / (lw - lb);
+        MAP3(a * powf(X + b, 2.4f));
+        return;
+    }
+    case PL_COLOR_TRC_GAMMA18: MAP3(powf(X, 1.8f)); goto scale_out;
+    case PL_COLOR_TRC_GAMMA20: MAP3(powf(X, 2.0f)); goto scale_out;
+    case PL_COLOR_TRC_UNKNOWN:
+    case PL_COLOR_TRC_GAMMA22: MAP3(powf(X, 2.2f)); goto scale_out;
+    case PL_COLOR_TRC_GAMMA24: MAP3(powf(X, 2.4f)); goto scale_out;
+    case PL_COLOR_TRC_GAMMA26: MAP3(powf(X, 2.6f)); goto scale_out;
+    case PL_COLOR_TRC_GAMMA28: MAP3(powf(X, 2.8f)); goto scale_out;
+    case PL_COLOR_TRC_PRO_PHOTO:
+        MAP3(X > 0.03125f ? powf(X, 1.8f) : X / 16);
+        goto scale_out;
+    case PL_COLOR_TRC_ST428:
+        MAP3(52.37f/48 * powf(X, 2.6f));
+        goto scale_out;
+    case PL_COLOR_TRC_PQ:
+        MAP3(powf(X, 1 / PQ_M2));
+        MAP3(fmaxf(X - PQ_C1, 0) / (PQ_C2 - PQ_C3 * X));
+        MAP3(10000 / PL_COLOR_SDR_WHITE * powf(X, 1 / PQ_M1));
+        return;
+    case PL_COLOR_TRC_HLG: {
+        const float y = fmaxf(1.2f + 0.42f * log10f(csp_max / HLG_REF), 1);
+        const float b = sqrtf(3 * powf(csp_min / csp_max, 1 / y));
+        const pl_matrix3x3 rgb2xyz =
+            pl_get_rgb2xyz_matrix(pl_raw_primaries_get(csp->primaries));
+        const float *coef = rgb2xyz.m[1];
+        // OETF^-1
+        MAP3((1 - b) * X + b);
+        MAP3(X > 0.5f ? expf((X - HLG_C) / HLG_A) + HLG_B
+                      : 4 * X * X);
+        // OOTF
+        float luma = coef[0] * color[0] + coef[1] * color[1] + coef[2] * color[2];
+        luma = powf(fmaxf(luma / 12, 0), y - 1);
+        MAP3(luma * X / 12);
+        return;
+    }
+    case PL_COLOR_TRC_V_LOG:
+        MAP3(X >= 0.181f ? powf(10, (X - VLOG_D) / VLOG_C) - VLOG_B
+                         : (X - 0.125f) / 5.6f);
+        return;
+    case PL_COLOR_TRC_S_LOG1:
+        MAP3(powf(10, (X - SLOG_C) / SLOG_A) - SLOG_B);
+        return;
+    case PL_COLOR_TRC_S_LOG2:
+        MAP3(X >= SLOG_Q ? (powf(10, (X - SLOG_C) / SLOG_A) - SLOG_B) / SLOG_K2
+                         : (X - SLOG_Q) / SLOG_P);
+        return;
+    case PL_COLOR_TRC_LINEAR:
+    case PL_COLOR_TRC_COUNT:
+        break;
+    }
+
+    pl_unreachable();
+
+scale_out:
+    MAP3((csp_max - csp_min) * X + csp_min);
+}
+
+void pl_color_delinearize(const struct pl_color_space *csp, float color[3])
+{
+    if (csp->transfer == PL_COLOR_TRC_LINEAR)
+        return;
+
+    float csp_min, csp_max;
+    pl_color_space_nominal_luma_ex(pl_nominal_luma_params(
+        .color      = csp,
+        .metadata   = PL_HDR_METADATA_HDR10,
+        .scaling    = PL_HDR_NORM,
+        .out_min    = &csp_min,
+        .out_max    = &csp_max,
+    ));
+
+    if (pl_color_space_is_black_scaled(csp) && csp->transfer != PL_COLOR_TRC_HLG)
+        MAP3((X - csp_min) / (csp_max - csp_min));
+
+    MAP3(fmaxf(X, 0));
+
+    switch (csp->transfer) {
+    case PL_COLOR_TRC_SRGB:
+        MAP3(X >= 0.0031308f ? 1.055f * powf(X, 1/2.4f) - 0.055f
+                             : 12.92f * X);
+        return;
+    case PL_COLOR_TRC_BT_1886: {
+        const float lb = powf(csp_min, 1/2.4f);
+        const float lw = powf(csp_max, 1/2.4f);
+        const float a = powf(lw - lb, 2.4f);
+        const float b = lb / (lw - lb);
+        MAP3(powf(X / a, 1/2.4f) - b);
+            return;
+    }
+    case PL_COLOR_TRC_GAMMA18: MAP3(powf(X, 1/1.8f)); return;
+    case PL_COLOR_TRC_GAMMA20: MAP3(powf(X, 1/2.0f)); return;
+    case PL_COLOR_TRC_UNKNOWN:
+    case PL_COLOR_TRC_GAMMA22: MAP3(powf(X, 1/2.2f)); return;
+    case PL_COLOR_TRC_GAMMA24: MAP3(powf(X, 1/2.4f)); return;
+    case PL_COLOR_TRC_GAMMA26: MAP3(powf(X, 1/2.6f)); return;
+    case PL_COLOR_TRC_GAMMA28: MAP3(powf(X, 1/2.8f)); return;
+    case PL_COLOR_TRC_ST428:
+        MAP3(powf(X * 48/52.37f, 1/2.6f));
+        return;
+    case PL_COLOR_TRC_PRO_PHOTO:
+        MAP3(X >= 0.001953f ? powf(X, 1/1.8f) : 16 * X);
+        return;
+    case PL_COLOR_TRC_PQ:
+        MAP3(powf(X * PL_COLOR_SDR_WHITE / 10000, PQ_M1));
+        MAP3(powf((PQ_C1 + PQ_C2 * X) / (1 + PQ_C3 * X), PQ_M2));
+        return;
+    case PL_COLOR_TRC_HLG: {
+        const float y = fmaxf(1.2f + 0.42f * log10f(csp_max / HLG_REF), 1);
+        const float b = sqrtf(3 * powf(csp_min / csp_max, 1 / y));
+        const pl_matrix3x3 rgb2xyz =
+            pl_get_rgb2xyz_matrix(pl_raw_primaries_get(csp->primaries));
+        const float *coef = rgb2xyz.m[1];
+        // OOTF^-1
+        float luma = coef[0] * color[0] + coef[1] * color[1] + coef[2] * color[2];
+        luma = fmaxf(1e-6f, powf(luma / csp_max, (1 - y) / y));
+        MAP3(12 / csp_max * luma * X);
+        // OETF
+        MAP3(X > 1 ? HLG_A * logf(X - HLG_B) + HLG_C : 0.5f * sqrtf(X));
+        MAP3((X - b) / (1 - b));
+        return;
+    }
+    case PL_COLOR_TRC_V_LOG:
+        MAP3(X >= 0.01f ? VLOG_C * log10f(X + VLOG_B) + VLOG_D
+                        : 5.6f * X + 0.125f);
+        return;
+    case PL_COLOR_TRC_S_LOG1:
+        MAP3(SLOG_A * log10f(X + SLOG_B) + SLOG_C);
+        return;
+    case PL_COLOR_TRC_S_LOG2:
+        MAP3(X >= 0 ? SLOG_A * log10f(SLOG_B * X + SLOG_C)
+                    : SLOG_P * X + SLOG_Q);
+        return;
+    case PL_COLOR_TRC_LINEAR:
+    case PL_COLOR_TRC_COUNT:
+        break;
+    }
+
+    pl_unreachable();
+}
+
 void pl_color_space_merge(struct pl_color_space *orig,
                           const struct pl_color_space *new)
 {
@@ -591,18 +838,6 @@ void pl_color_space_nominal_luma_ex(const struct pl_nominal_luma_params *params)
         *params->out_avg = avg_luma;
 }
 
-void pl_color_space_nominal_luma(const struct pl_color_space *csp,
-                                 float *out_min, float *out_max)
-{
-    pl_color_space_nominal_luma_ex(pl_nominal_luma_params(
-        .color      = csp,
-        .metadata   = PL_HDR_METADATA_ANY,
-        .scaling    = PL_HDR_NORM,
-        .out_min    = out_min,
-        .out_max    = out_max,
-    ));
-}
-
 void pl_color_space_infer(struct pl_color_space *space)
 {
     if (!space->primaries)
@@ -690,27 +925,33 @@ void pl_color_space_infer_ref(struct pl_color_space *space,
 void pl_color_space_infer_map(struct pl_color_space *src,
                               struct pl_color_space *dst)
 {
-    bool unknown_contrast = !src->hdr.min_luma;
+    bool unknown_src_contrast = !src->hdr.min_luma;
+    bool unknown_dst_contrast = !dst->hdr.min_luma;
 
     infer_both_ref(dst, src);
 
-    // If the src has an unspecified SDR-like gamma, default it to match the
-    // dst colorspace contrast. This does not matter in most cases, but ensures
-    // that BT.1886 is tuned to the appropriate black point by default.
-    bool dynamic_contrast = pl_color_space_is_black_scaled(src) ||
-                            src->transfer == PL_COLOR_TRC_BT_1886;
-
-    if (unknown_contrast && dynamic_contrast)
+    // If the src has an unspecified gamma curve with dynamic black scaling,
+    // default it to match the dst colorspace contrast. This does not matter in
+    // most cases, but ensures that BT.1886 is tuned to the appropriate black
+    // point by default.
+    bool dynamic_src_contrast = pl_color_space_is_black_scaled(src) ||
+                                src->transfer == PL_COLOR_TRC_BT_1886;
+    if (unknown_src_contrast && dynamic_src_contrast)
         src->hdr.min_luma = dst->hdr.min_luma;
+
+    // Do the same in reverse if both src and dst are SDR curves
+    bool src_is_sdr = !pl_color_space_is_hdr(src);
+    bool dst_is_sdr = !pl_color_space_is_hdr(dst);
+    if (unknown_dst_contrast && src_is_sdr && dst_is_sdr)
+        dst->hdr.min_luma = src->hdr.min_luma;
+
+    // If the src is HLG and the output is HDR, tune the HLG peak to the output
+    if (src->transfer == PL_COLOR_TRC_HLG && pl_color_space_is_hdr(dst))
+        src->hdr.max_luma = dst->hdr.max_luma;
 }
 
 const struct pl_color_adjustment pl_color_adjustment_neutral = {
-    .brightness     = 0.0,
-    .contrast       = 1.0,
-    .saturation     = 1.0,
-    .hue            = 0.0,
-    .gamma          = 1.0,
-    .temperature    = 0.0,
+    PL_COLOR_ADJUSTMENT_NEUTRAL
 };
 
 void pl_chroma_location_offset(enum pl_chroma_location loc, float *x, float *y)
@@ -948,16 +1189,15 @@ pl_matrix3x3 pl_get_rgb2xyz_matrix(const struct pl_raw_primaries *prim)
     pl_matrix3x3 out = {{{0}}};
     float S[3], X[4], Z[4];
 
-    // Convert from CIE xyY to XYZ. Note that Y=1 holds true for all primaries
-    X[0] = prim->red.x   / prim->red.y;
-    X[1] = prim->green.x / prim->green.y;
-    X[2] = prim->blue.x  / prim->blue.y;
-    X[3] = prim->white.x / prim->white.y;
+    X[0] = pl_cie_X(prim->red);
+    X[1] = pl_cie_X(prim->green);
+    X[2] = pl_cie_X(prim->blue);
+    X[3] = pl_cie_X(prim->white);
 
-    Z[0] = (1 - prim->red.x   - prim->red.y)   / prim->red.y;
-    Z[1] = (1 - prim->green.x - prim->green.y) / prim->green.y;
-    Z[2] = (1 - prim->blue.x  - prim->blue.y)  / prim->blue.y;
-    Z[3] = (1 - prim->white.x - prim->white.y) / prim->white.y;
+    Z[0] = pl_cie_Z(prim->red);
+    Z[1] = pl_cie_Z(prim->green);
+    Z[2] = pl_cie_Z(prim->blue);
+    Z[3] = pl_cie_Z(prim->white);
 
     // S = XYZ^-1 * W
     for (int i = 0; i < 3; i++) {
@@ -989,12 +1229,11 @@ pl_matrix3x3 pl_get_xyz2rgb_matrix(const struct pl_raw_primaries *prim)
     return out;
 }
 
-// LMS<-XYZ revised matrix from CIECAM97, based on a linear transform and
-// normalized for equal energy on monochrome inputs
-static const pl_matrix3x3 m_cat97 = {{
-    {  0.8562,  0.3372, -0.1934 },
-    { -0.8360,  1.8327,  0.0033 },
-    {  0.0357, -0.0469,  1.0112 },
+// Matrix used in CAT16, a revised one-step linear transform method
+static const pl_matrix3x3 m_cat16 = {{
+    {  0.401288, 0.650173, -0.051461 },
+    { -0.250268, 1.204414,  0.045854 },
+    { -0.002079, 0.048952,  0.953127 },
 }};
 
 // M := M * XYZd<-XYZs
@@ -1007,21 +1246,20 @@ static void apply_chromatic_adaptation(struct pl_cie_xy src,
     if (fabs(src.x - dest.x) < 1e-6 && fabs(src.y - dest.y) < 1e-6)
         return;
 
-    // XYZd<-XYZs = Ma^-1 * (I*[Cd/Cs]) * Ma
+    // Linear "von Kries" method, adapted from CIECAM16
     // http://www.brucelindbloom.com/index.html?Eqn_ChromAdapt.html
-    // For Ma, we use the CIECAM97 revised (linear) matrix
     float C[3][2];
 
     for (int i = 0; i < 3; i++) {
         // source cone
-        C[i][0] = m_cat97.m[i][0] * pl_cie_X(src)
-                + m_cat97.m[i][1] * 1
-                + m_cat97.m[i][2] * pl_cie_Z(src);
+        C[i][0] = m_cat16.m[i][0] * pl_cie_X(src)
+                + m_cat16.m[i][1] * 1
+                + m_cat16.m[i][2] * pl_cie_Z(src);
 
         // dest cone
-        C[i][1] = m_cat97.m[i][0] * pl_cie_X(dest)
-                + m_cat97.m[i][1] * 1
-                + m_cat97.m[i][2] * pl_cie_Z(dest);
+        C[i][1] = m_cat16.m[i][0] * pl_cie_X(dest)
+                + m_cat16.m[i][1] * 1
+                + m_cat16.m[i][2] * pl_cie_Z(dest);
     }
 
     // tmp := I * [Cd/Cs] * Ma
@@ -1029,10 +1267,10 @@ static void apply_chromatic_adaptation(struct pl_cie_xy src,
     for (int i = 0; i < 3; i++)
         tmp.m[i][i] = C[i][1] / C[i][0];
 
-    pl_matrix3x3_mul(&tmp, &m_cat97);
+    pl_matrix3x3_mul(&tmp, &m_cat16);
 
     // M := M * Ma^-1 * tmp
-    pl_matrix3x3 ma_inv = m_cat97;
+    pl_matrix3x3 ma_inv = m_cat16;
     pl_matrix3x3_invert(&ma_inv);
     pl_matrix3x3_mul(mat, &ma_inv);
     pl_matrix3x3_mul(mat, &tmp);
@@ -1114,7 +1352,7 @@ pl_matrix3x3 pl_get_cone_matrix(const struct pl_cone_params *params,
                                 const struct pl_raw_primaries *prim)
 {
     // LMS<-RGB := LMS<-XYZ * XYZ<-RGB
-    pl_matrix3x3 rgb2lms = m_cat97;
+    pl_matrix3x3 rgb2lms = m_cat16;
     pl_matrix3x3 rgb2xyz = pl_get_rgb2xyz_matrix(prim);
     pl_matrix3x3_mul(&rgb2lms, &rgb2xyz);
 
@@ -1288,8 +1526,8 @@ static float test_point_gamut(struct pl_cie_xy point,
           d2 = test_point_line(point, prim->green, prim->blue),
           d3 = test_point_line(point, prim->blue, prim->red);
 
-    bool has_neg = d1 < 0 || d2 < 0 || d3 < 0,
-         has_pos = d1 > 0 || d2 > 0 || d3 > 0;
+    bool has_neg = d1 < -1e-6f || d2 < -1e-6f || d3 < -1e-6f,
+         has_pos = d1 >  1e-6f || d2 >  1e-6f || d3 >  1e-6f;
 
     return !(has_neg && has_pos);
 }
@@ -1311,6 +1549,65 @@ bool pl_primaries_valid(const struct pl_raw_primaries *prim)
     return fabs(area) > 1e-6 && test_point_gamut(prim->white, prim);
 }
 
+static inline float xy_dist2(struct pl_cie_xy a, struct pl_cie_xy b)
+{
+    const float dx = a.x - b.x, dy = a.y - b.y;
+    return dx * dx + dy * dy;
+}
+
+bool pl_primaries_compatible(const struct pl_raw_primaries *a,
+                             const struct pl_raw_primaries *b)
+{
+    float RR = xy_dist2(a->red, b->red),    RG = xy_dist2(a->red, b->green),
+          RB = xy_dist2(a->red, b->blue),   GG = xy_dist2(a->green, b->green),
+          GB = xy_dist2(a->green, b->blue), BB = xy_dist2(a->blue, b->blue);
+    return RR < RG && RR < RB && GG < RG && GG < GB && BB < RB && BB < GB;
+}
+
+// returns the intersection of the two lines defined by ab and cd
+static struct pl_cie_xy intersection(struct pl_cie_xy a, struct pl_cie_xy b,
+                                     struct pl_cie_xy c, struct pl_cie_xy d)
+{
+    float det = (a.x - b.x) * (c.y - d.y) - (a.y - b.y) * (c.x - d.x);
+    float t = ((a.x - c.x) * (c.y - d.y) - (a.y - c.y) * (c.x - d.x)) / det;
+    return (struct pl_cie_xy) {
+        .x = t ? a.x + t * (b.x - a.x) : 0.0f,
+        .y = t ? a.y + t * (b.y - a.y) : 0.0f,
+    };
+}
+
+// x, y, z specified in clockwise order, with a, b, c being the enclosing gamut
+static struct pl_cie_xy
+clip_point(struct pl_cie_xy x, struct pl_cie_xy y, struct pl_cie_xy z,
+           struct pl_cie_xy a, struct pl_cie_xy b, struct pl_cie_xy c)
+{
+    const float d1 = test_point_line(y, a, b);
+    const float d2 = test_point_line(y, b, c);
+    if (d1 <= 0.0f && d2 <= 0.0f) {
+        return y; // already inside triangle
+    } else if (d1 > 0.0f && d2 > 0.0f) {
+        return b; // target vertex fully enclosed
+    } else if (d1 > 0.0f) {
+        return intersection(a, b, y, z);
+    } else {
+        return intersection(x, y, b, c);
+    }
+}
+
+struct pl_raw_primaries pl_primaries_clip(const struct pl_raw_primaries *src,
+                                          const struct pl_raw_primaries *dst)
+{
+    return (struct pl_raw_primaries) {
+        .red   = clip_point(src->green, src->red, src->blue,
+                            dst->green, dst->red, dst->blue),
+        .green = clip_point(src->blue, src->green, src->red,
+                            dst->blue, dst->green, dst->red),
+        .blue  = clip_point(src->red, src->blue, src->green,
+                            dst->red, dst->blue, dst->green),
+        .white = src->white,
+    };
+}
+
 /* Fill in the Y, U, V vectors of a yuv-to-rgb conversion matrix
  * based on the given luma weights of the R, G and B components (lr, lg, lb).
  * lr+lg+lb is assumed to equal 1.
@@ -1534,11 +1831,11 @@ bool pl_icc_profile_equal(const struct pl_icc_profile *p1,
 
 void pl_icc_profile_compute_signature(struct pl_icc_profile *profile)
 {
+    if (!profile->len)
+        profile->signature = 0;
+
     // In theory, we could get this value from the profile header itself if
     // lcms is available, but I'm not sure if it's even worth the trouble. Just
-    // hard-code this to a siphash64(), which is decently fast anyway.
-    profile->signature = pl_str_hash((pl_str) {
-        .buf = (uint8_t *) profile->data,
-        .len = profile->len
-    });
+    // hard-code this to a pl_mem_hash(), which is decently fast anyway.
+    profile->signature = pl_mem_hash(profile->data, profile->len);
 }
