@@ -36,6 +36,20 @@ pl_dispatch pl_gpu_dispatch(pl_gpu gpu)
     return impl->dp;
 }
 
+pl_cache pl_gpu_cache(pl_gpu gpu)
+{
+    if (!gpu)
+        return NULL;
+    const struct pl_gpu_fns *impl = PL_PRIV(gpu);
+    return atomic_load(&impl->cache);
+}
+
+void pl_gpu_set_cache(pl_gpu gpu, pl_cache cache)
+{
+    struct pl_gpu_fns *impl = PL_PRIV(gpu);
+    atomic_store(&impl->cache, cache);
+}
+
 bool pl_fmt_is_ordered(pl_fmt fmt)
 {
     bool ret = !fmt->opaque;
@@ -174,6 +188,7 @@ static inline bool check_mod(pl_gpu gpu, pl_fmt fmt, uint64_t mod)
 
 pl_tex pl_tex_create(pl_gpu gpu, const struct pl_tex_params *params)
 {
+    require(params->format);
     require(!params->import_handle || !params->export_handle);
     require(!params->import_handle || !params->initial_data);
     if (params->export_handle) {
@@ -202,6 +217,7 @@ pl_tex pl_tex_create(pl_gpu gpu, const struct pl_tex_params *params)
         require(!params->renderable);
         require(!params->blit_src || gpu->limits.blittable_1d_3d);
         require(!params->blit_dst || gpu->limits.blittable_1d_3d);
+        require(!params->format->num_planes);
         break;
     case 2:
         require(params->w > 0 && params->h > 0);
@@ -216,10 +232,10 @@ pl_tex pl_tex_create(pl_gpu gpu, const struct pl_tex_params *params)
         require(!params->renderable);
         require(!params->blit_src || gpu->limits.blittable_1d_3d);
         require(!params->blit_dst || gpu->limits.blittable_1d_3d);
+        require(!params->format->num_planes);
         break;
     }
 
-    require(params->format);
     enum pl_fmt_caps fmt_caps = params->format->caps;
     bool fmt_opaque = params->format->opaque;
     for (int i = 0; i < params->format->num_planes; i++) {
@@ -434,6 +450,8 @@ static bool fix_tex_transfer(pl_gpu gpu, struct pl_tex_transfer_params *params)
     if (!params->depth_pitch || !tex->params.d)
         params->depth_pitch = pl_rect_h(rc) * params->row_pitch;
 
+    require(params->row_pitch);
+    require(params->depth_pitch);
     params->rc = rc;
 
     // Check the parameters for sanity
@@ -572,7 +590,11 @@ pl_buf pl_buf_create(pl_gpu gpu, const struct pl_buf_params *params)
     require(!params->uniform || params->size <= gpu->limits.max_ubo_size);
     require(!params->storable || params->size <= gpu->limits.max_ssbo_size);
     require(!params->drawable || params->size <= gpu->limits.max_vbo_size);
-    require(!params->host_mapped || params->size <= gpu->limits.max_mapped_size);
+    if (params->host_mapped) {
+        require(params->size <= gpu->limits.max_mapped_size);
+        require(params->memory_type != PL_BUF_MEM_DEVICE ||
+                params->size <= gpu->limits.max_mapped_vram);
+    }
 
     if (params->format) {
         pl_fmt fmt = params->format;
@@ -1255,42 +1277,6 @@ bool pl_gpu_is_failed(pl_gpu gpu)
     return impl->gpu_is_failed(gpu);
 }
 
-pl_sync pl_sync_create(pl_gpu gpu, enum pl_handle_type handle_type)
-{
-    require(handle_type);
-    require(handle_type & gpu->export_caps.sync);
-    require(PL_ISPOT(handle_type));
-
-    const struct pl_gpu_fns *impl = PL_PRIV(gpu);
-    return impl->sync_create(gpu, handle_type);
-
-error:
-    return NULL;
-}
-
-void pl_sync_destroy(pl_gpu gpu, pl_sync *sync)
-{
-    if (!*sync)
-        return;
-
-    const struct pl_gpu_fns *impl = PL_PRIV(gpu);
-    impl->sync_destroy(gpu, *sync);
-    *sync = NULL;
-}
-
-bool pl_tex_export(pl_gpu gpu, pl_tex tex, pl_sync sync)
-{
-    require(tex->params.import_handle || tex->params.export_handle);
-
-    const struct pl_gpu_fns *impl = PL_PRIV(gpu);
-    return impl->tex_export(gpu, tex, sync);
-
-error:
-    if (tex->params.debug_tag)
-        PL_ERR(gpu, "  for texture: %s", tex->params.debug_tag);
-    return false;
-}
-
 pl_timer pl_timer_create(pl_gpu gpu)
 {
     const struct pl_gpu_fns *impl = PL_PRIV(gpu);
