@@ -155,8 +155,8 @@ struct pl_vulkan_t {
     // form. This list does not contain duplicates, nor any extra queues
     // enabled at device creation time. Deprecated in favor of querying
     // `vkGetPhysicalDeviceQueueFamilyProperties` directly.
-    const struct pl_vulkan_queue *queues PL_DEPRECATED;
-    int num_queues PL_DEPRECATED;
+    PL_DEPRECATED_IN(v6.271) const struct pl_vulkan_queue *queues;
+    PL_DEPRECATED_IN(v6.271) int num_queues;
 };
 
 struct pl_vulkan_params {
@@ -271,9 +271,8 @@ struct pl_vulkan_params {
 PL_API extern const struct pl_vulkan_params pl_vulkan_default_params;
 
 // Creates a new vulkan device based on the given parameters and initializes
-// a new GPU. This function will internally initialize a VkDevice. There is
-// currently no way to share a vulkan device with the caller. If `params` is
-// left as NULL, it defaults to &pl_vulkan_default_params.
+// a new GPU. If `params` is left as NULL, it defaults to
+// &pl_vulkan_default_params.
 //
 // Thread-safety: Safe
 PL_API pl_vulkan pl_vulkan_create(pl_log log, const struct pl_vulkan_params *params);
@@ -347,6 +346,11 @@ struct pl_vulkan_swapchain_params {
     // calling `pl_swapchain_resize` as appropriate. libplacebo will tolerate
     // the "suboptimal" status indefinitely.
     bool allow_suboptimal;
+
+    // Disable high-bit (10 or more) SDR formats. May help work around buggy
+    // drivers which don't dither properly when outputting high bit depth
+    // SDR backbuffers to 8-bit screens.
+    bool disable_10bit_sdr;
 };
 
 #define pl_vulkan_swapchain_params(...) (&(struct pl_vulkan_swapchain_params) { __VA_ARGS__ })
@@ -491,7 +495,7 @@ struct pl_vulkan_wrap_params {
 
 // Wraps an external VkImage into a pl_tex abstraction. By default, the image
 // is considered "held" by the user and must be released before calling any
-// pl_tex_* API calls on it (see `pl_vulkan_release`).
+// pl_tex_* API calls on it (see `pl_vulkan_release_ex`).
 //
 // This wrapper can be destroyed by simply calling `pl_tex_destroy` on it,
 // which will not destroy the underlying VkImage. If a pl_tex wrapper is
@@ -509,7 +513,7 @@ PL_API pl_tex pl_vulkan_wrap(pl_gpu gpu, const struct pl_vulkan_wrap_params *par
 // ones created by `pl_tex_create`) and unwraps it to expose the underlying
 // VkImage to the user. Unlike `pl_vulkan_wrap`, this `pl_tex` is *not*
 // considered held after calling this function - the user must explicitly
-// `pl_vulkan_hold` before accessing the VkImage.
+// `pl_vulkan_hold_ex` before accessing the VkImage.
 //
 // `out_format` and `out_flags` will be updated to hold the VkImage's
 // format and usage flags. (Optional)
@@ -621,14 +625,6 @@ struct pl_vulkan_sem_params {
 PL_API VkSemaphore pl_vulkan_sem_create(pl_gpu gpu, const struct pl_vulkan_sem_params *params);
 PL_API void pl_vulkan_sem_destroy(pl_gpu gpu, VkSemaphore *semaphore);
 
-// Backwards-compatibility wrappers for older versions of the API.
-PL_DEPRECATED PL_API bool pl_vulkan_hold(pl_gpu gpu, pl_tex tex, VkImageLayout layout,
-                                         pl_vulkan_sem sem_out);
-PL_DEPRECATED PL_API bool pl_vulkan_hold_raw(pl_gpu gpu, pl_tex tex, VkImageLayout *out_layout,
-                                             pl_vulkan_sem sem_out);
-PL_DEPRECATED PL_API void pl_vulkan_release(pl_gpu gpu, pl_tex tex, VkImageLayout layout,
-                                            pl_vulkan_sem sem_in);
-
 PL_API_END
 
 #endif // LIBPLACEBO_VULKAN_H_
