@@ -20,12 +20,35 @@
 #include "../common.h"
 #include "../log.h"
 #include "../gpu.h"
+#include "pl_thread.h"
 
-#include <epoxy/gl.h>
+#include <libplacebo/opengl.h>
 
-#ifdef EPOXY_HAS_EGL
-#include <epoxy/egl.h>
-#endif
+// Collision with llvm-mingw <winnt.h>
+#undef MemoryBarrier
+
+#define GLAD_GL
+#define GLAD_GLES2
+#include <glad/gl.h>
+#include <glad/egl.h>
+
+typedef GladGLContext gl_funcs;
+
+// PL_PRIV(pl_opengl)
+struct gl_ctx {
+    pl_log log;
+    struct pl_opengl_params params;
+    bool is_debug;
+    bool is_debug_egl;
+    bool is_gles;
+
+    // For context locking
+    pl_mutex lock;
+    int count;
+
+    // Dispatch table
+    gl_funcs func;
+};
 
 struct gl_cb {
     void (*callback)(void *priv);
