@@ -49,6 +49,7 @@ enum {
     I16   = GL_SHORT,
     I32   = GL_INT,
     FLT   = GL_FLOAT,
+    HALF  = GL_HALF_FLOAT,
 
     // Component aliases
     R     = GL_RED,
@@ -76,6 +77,16 @@ const struct gl_format formats_norm8[] = {
     {GL_RGBA8,          RGBA,  U8,  FMT("rgba8",    8, UNORM, S|L|F|V)},
 };
 
+// Signed variants
+/* TODO: these are broken in mesa
+const struct gl_format formats_snorm8[] = {
+    {GL_R8_SNORM,       R,     I8,  FMT("r8s",      8, SNORM, S|L|F|V)},
+    {GL_RG8_SNORM,      RG,    I8,  FMT("rg8s",     8, SNORM, S|L|F|V)},
+    {GL_RGB8_SNORM,     RGB,   I8,  FMT("rgb8s",    8, SNORM, S|L|F|V)},
+    {GL_RGBA8_SNORM,    RGBA,  I8,  FMT("rgba8s",   8, SNORM, S|L|F|V)},
+};
+*/
+
 // BGRA 8-bit
 const struct gl_format formats_bgra8[] = {
     {GL_RGBA8,          BGRA,  U8,  {
@@ -104,32 +115,62 @@ const struct gl_format formats_rgb16_fallback[] = {
     {GL_RGB16,          RGB,   U16, FMT("rgb16",   16, UNORM, S|L|V)},
 };
 
-// Floating point texture formats
-const struct gl_format formats_float[] = {
-    {GL_R16F,           R,     FLT, FMT("r16f",    16, FLOAT, S|L|F)},
-    {GL_RG16F,          RG,    FLT, FMT("rg16f",   16, FLOAT, S|L|F)},
-    {GL_RGB16F,         RGB,   FLT, FMT("rgb16f",  16, FLOAT, S|L|F)},
-    {GL_RGBA16F,        RGBA,  FLT, FMT("rgba16f", 16, FLOAT, S|L|F)},
+// Signed 16-bit variants
+/* TODO: these are broken in mesa and nvidia
+const struct gl_format formats_snorm16[] = {
+    {GL_R16_SNORM,      R,     I16, FMT("r16s",    16, SNORM, S|L|F|V)},
+    {GL_RG16_SNORM,     RG,    I16, FMT("rg16s",   16, SNORM, S|L|F|V)},
+    {GL_RGB16_SNORM,    RGB,   I16, FMT("rgb16s",  16, SNORM, S|L|F|V)},
+    {GL_RGBA16_SNORM,   RGBA,  I16, FMT("rgba16s", 16, SNORM, S|L|F|V)},
+};
+*/
+
+// 32-bit floating point texture formats
+const struct gl_format formats_float32[] = {
     {GL_R32F,           R,     FLT, FMT("r32f",    32, FLOAT, S|L|F|V)},
     {GL_RG32F,          RG,    FLT, FMT("rg32f",   32, FLOAT, S|L|F|V)},
     {GL_RGB32F,         RGB,   FLT, FMT("rgb32f",  32, FLOAT, S|L|F|V)},
     {GL_RGBA32F,        RGBA,  FLT, FMT("rgba32f", 32, FLOAT, S|L|F|V)},
 };
 
+// 16-bit floating point texture formats
+const struct gl_format formats_float16[] = {
+    {GL_R16F,           R,     FLT,  FMT("r16f",    16, FLOAT, S|L|F)},
+    {GL_RG16F,          RG,    FLT,  FMT("rg16f",   16, FLOAT, S|L|F)},
+    {GL_RGB16F,         RGB,   FLT,  FMT("rgb16f",  16, FLOAT, S|L|F)},
+    {GL_RGBA16F,        RGBA,  FLT,  FMT("rgba16f", 16, FLOAT, S|L|F)},
+};
+
+// 16-bit half float texture formats
+const struct gl_format formats_half16[] = {
+    {GL_R16F,           R,     HALF, FMT("r16hf",   16, FLOAT, S|L|F)},
+    {GL_RG16F,          RG,    HALF, FMT("rg16hf",  16, FLOAT, S|L|F)},
+    {GL_RGB16F,         RGB,   HALF, FMT("rgb16hf", 16, FLOAT, S|L|F)},
+    {GL_RGBA16F,        RGBA,  HALF, FMT("rgba16hf",16, FLOAT, S|L|F)},
+};
+
 // Renderable 16-bit float formats (excluding rgb16f)
 const struct gl_format formats_float16_fbo[] = {
-    {GL_R16F,           R,     FLT, FMT("r16f",    16, FLOAT, S|L|F)},
-    {GL_RG16F,          RG,    FLT, FMT("rg16f",   16, FLOAT, S|L|F)},
-    {GL_RGB16F,         RGB,   FLT, FMT("rgb16f",  16, FLOAT, S|L)},
-    {GL_RGBA16F,        RGBA,  FLT, FMT("rgba16f", 16, FLOAT, S|L|F)},
+    {GL_R16F,           R,     HALF, FMT("r16hf",   16, FLOAT, S|L|F)},
+    {GL_RG16F,          RG,    HALF, FMT("rg16hf",  16, FLOAT, S|L|F)},
+    {GL_RGB16F,         RGB,   HALF, FMT("rgb16hf", 16, FLOAT, S|L)},
+    {GL_RGBA16F,        RGBA,  HALF, FMT("rgba16hf",16, FLOAT, S|L|F)},
+    {GL_R16F,           R,     FLT,  FMT("r16f",    16, FLOAT, S|L|F)},
+    {GL_RG16F,          RG,    FLT,  FMT("rg16f",   16, FLOAT, S|L|F)},
+    {GL_RGB16F,         RGB,   FLT,  FMT("rgb16f",  16, FLOAT, S|L)},
+    {GL_RGBA16F,        RGBA,  FLT,  FMT("rgba16f", 16, FLOAT, S|L|F)},
 };
 
 // Non-renderable 16-bit float formats
 const struct gl_format formats_float16_fallback[] = {
-    {GL_R16F,           R,     FLT, FMT("r16f",    16, FLOAT, S|L)},
-    {GL_RG16F,          RG,    FLT, FMT("rg16f",   16, FLOAT, S|L)},
-    {GL_RGB16F,         RGB,   FLT, FMT("rgb16f",  16, FLOAT, S|L)},
-    {GL_RGBA16F,        RGBA,  FLT, FMT("rgba16f", 16, FLOAT, S|L)},
+    {GL_R16F,           R,     HALF, FMT("r16hf",   16, FLOAT, S|L)},
+    {GL_RG16F,          RG,    HALF, FMT("rg16hf",  16, FLOAT, S|L)},
+    {GL_RGB16F,         RGB,   HALF, FMT("rgb16hf", 16, FLOAT, S|L)},
+    {GL_RGBA16F,        RGBA,  HALF, FMT("rgba16hf",16, FLOAT, S|L)},
+    {GL_R16F,           R,     FLT,  FMT("r16f",    16, FLOAT, S|L)},
+    {GL_RG16F,          RG,    FLT,  FMT("rg16f",   16, FLOAT, S|L)},
+    {GL_RGB16F,         RGB,   FLT,  FMT("rgb16f",  16, FLOAT, S|L)},
+    {GL_RGBA16F,        RGBA,  FLT,  FMT("rgba16f", 16, FLOAT, S|L)},
 };
 
 // (Unsigned) integer formats
@@ -150,15 +191,6 @@ const struct gl_format formats_uint[] = {
     {GL_RGB32UI,        RGBI,  U32, FMT("rgb32u",  32, UINT)},
     {GL_RGBA32UI,       RGBAI, U32, FMT("rgba32u", 32, UINT)},
 
-    {GL_R8_SNORM,       R,     I8,  FMT("r8s",      8, SNORM)},
-    {GL_RG8_SNORM,      RG,    I8,  FMT("rg8s",     8, SNORM)},
-    {GL_RGB8_SNORM,     RGB,   I8,  FMT("rgb8s",    8, SNORM)},
-    {GL_RGBA8_SNORM,    RGBA,  I8,  FMT("rgba8s",   8, SNORM)},
-    {GL_R16_SNORM,      R,     I16, FMT("r16s",    16, SNORM)},
-    {GL_RG16_SNORM,     RG,    I16, FMT("rg16s",   16, SNORM)},
-    {GL_RGB16_SNORM,    RGB,   I16, FMT("rgb16s",  16, SNORM)},
-    {GL_RGBA16_SNORM,   RGBA,  I16, FMT("rgba16s", 16, SNORM)},
-
     {GL_R8I,            RI,    I8,  FMT("r8i",      8, SINT)},
     {GL_RG8I,           RGI,   I8,  FMT("rg8i",     8, SINT)},
     {GL_RGB8I,          RGBI,  I8,  FMT("rgb8i",    8, SINT)},
@@ -247,6 +279,7 @@ static void add_format(pl_gpu pgpu, const struct gl_format *gl_fmt)
         break;
     case GL_SHORT:
     case GL_UNSIGNED_SHORT:
+    case GL_HALF_FLOAT:
         size = 2;
         break;
     case GL_INT:
@@ -339,8 +372,17 @@ static void add_format(pl_gpu pgpu, const struct gl_format *gl_fmt)
     if (fmt->caps & PL_FMT_CAP_SAMPLEABLE)
         fmt->gatherable = p->gather_comps >= fmt->num_components;
 
+    bool host_readable = false;
+    if (p->gl_ver && p->has_readback)
+        host_readable = true;
     // Reading from textures on GLES requires FBO support for this fmt
-    if (p->has_readback && (p->gl_ver || (fmt->caps & PL_FMT_CAP_RENDERABLE)))
+    if (fmt->caps & PL_FMT_CAP_RENDERABLE) {
+        // this combination always works in glReadPixels
+        if ((gl_fmt->fmt == GL_RGBA && gl_fmt->type == GL_UNSIGNED_BYTE) ||
+            p->has_readback)
+            host_readable = true;
+    }
+    if (host_readable)
         fmt->caps |= PL_FMT_CAP_HOST_READABLE;
 
     if (gpu->glsl.compute && fmt->glsl_format && p->has_storage)
@@ -404,7 +446,9 @@ bool gl_setup_formats(struct pl_gpu_t *gpu)
         DO_FORMATS(formats_bgra8);
         DO_FORMATS(formats_norm16);
         DO_FORMATS(formats_rgb16_fbo);
-        DO_FORMATS(formats_float);
+        DO_FORMATS(formats_float32);
+        DO_FORMATS(formats_float16);
+        DO_FORMATS(formats_half16);
         DO_FORMATS(formats_uint);
         goto done;
     }
@@ -420,7 +464,12 @@ bool gl_setup_formats(struct pl_gpu_t *gpu)
             DO_FORMATS(formats_bgra8);
             DO_FORMATS(formats_norm16);
             DO_FORMATS(formats_rgb16_fbo);
-            DO_FORMATS(formats_float);
+            DO_FORMATS(formats_float32);
+            DO_FORMATS(formats_float16);
+            if (pl_opengl_has_ext(p->gl, "GL_ARB_half_float_pixel"))
+            {
+                DO_FORMATS(formats_half16);
+            }
         } else {
             // Fallback for GL2
             DO_FORMATS(formats_legacy_gl2);
@@ -439,11 +488,9 @@ bool gl_setup_formats(struct pl_gpu_t *gpu)
         }
         if (pl_opengl_has_ext(p->gl, "GL_EXT_texture_format_BGRA8888"))
             DO_FORMATS(formats_bgra_gles);
-        if (pl_opengl_has_ext(p->gl, "GL_EXT_texture_integer"))
-            DO_FORMATS(formats_uint);
+        DO_FORMATS(formats_uint);
         DO_FORMATS(formats_basic_vertex);
-        if (p->gles_ver >= 32 || pl_opengl_has_ext(p->gl, "GL_EXT_color_buffer_half_float")
-                              || pl_opengl_has_ext(p->gl, "GL_EXT_color_buffer_float")) {
+        if (p->gles_ver >= 32 || pl_opengl_has_ext(p->gl, "GL_EXT_color_buffer_float")) {
             DO_FORMATS(formats_float16_fbo);
         } else {
             DO_FORMATS(formats_float16_fallback);
