@@ -119,11 +119,13 @@ pl_gpu pl_gpu_create_gl(pl_log log, pl_opengl pl_gl, const struct pl_opengl_para
     p->gles_ver = glsl->gles ? ver : 0;
 
     // If possible, query the GLSL version from the implementation
-    const char *glslver = (char *) gl->GetString(GL_SHADING_LANGUAGE_VERSION);
-    if (glslver) {
-        PL_INFO(gpu, "    GL_SHADING_LANGUAGE_VERSION: %s", glslver);
+    const char *glslver_p = (char *) gl->GetString(GL_SHADING_LANGUAGE_VERSION);
+    pl_str glslver = pl_str0(glslver_p);
+    if (glslver.len) {
+        PL_INFO(gpu, "    GL_SHADING_LANGUAGE_VERSION: %.*s", PL_STR_FMT(glslver));
+        pl_str_eatstart0(&glslver, "OpenGL ES GLSL ES ");
         int major = 0, minor = 0;
-        if (sscanf(glslver, "%d.%d", &major, &minor) == 2)
+        if (pl_str_sscanf(glslver, "%d.%d", &major, &minor) == 2)
             glsl->version = major * 100 + minor;
     }
 
@@ -138,7 +140,21 @@ pl_gpu pl_gpu_create_gl(pl_log log, pl_opengl pl_gl, const struct pl_opengl_para
         }
     }
 
-    if (gl_test_ext(gpu, "GL_ARB_compute_shader", 43, 0)) {
+    static const int glsl_ver_req = 130;
+    if (glsl->version < glsl_ver_req) {
+        PL_FATAL(gpu, "GLSL version too old (%d < %d), please use a newer "
+                 "OpenGL implementation or downgrade libplacebo!",
+                 glsl->version, glsl_ver_req);
+        goto error;
+    }
+
+    if (params->max_glsl_version && params->max_glsl_version >= glsl_ver_req) {
+        glsl->version = PL_MIN(glsl->version, params->max_glsl_version);
+        PL_INFO(gpu, "Restricting GLSL version to %d... new version is %d",
+                params->max_glsl_version, glsl->version);
+    }
+
+    if (gl_test_ext(gpu, "GL_ARB_compute_shader", 43, 0) && glsl->version >= 420) {
         glsl->compute = true;
         get(GL_MAX_COMPUTE_SHARED_MEMORY_SIZE, &glsl->max_shmem_size);
         get(GL_MAX_COMPUTE_WORK_GROUP_INVOCATIONS, &glsl->max_group_threads);
@@ -146,8 +162,12 @@ pl_gpu pl_gpu_create_gl(pl_log log, pl_opengl pl_gl, const struct pl_opengl_para
             geti(GL_MAX_COMPUTE_WORK_GROUP_SIZE, i, &glsl->max_group_size[i]);
     }
 
-    if (gl_test_ext(gpu, "GL_ARB_texture_gather", 40, 0)) {
-        get(GL_MAX_PROGRAM_TEXTURE_GATHER_COMPONENTS_ARB, &p->gather_comps);
+    if (gl_test_ext(gpu, "GL_ARB_texture_gather", 40, 31) &&
+        glsl->version >= (p->gles_ver ? 310 : 400)) {
+        if (p->gles_ver)
+            p->gather_comps = 4;
+        else
+            get(GL_MAX_PROGRAM_TEXTURE_GATHER_COMPONENTS_ARB, &p->gather_comps);
         get(GL_MIN_PROGRAM_TEXTURE_GATHER_OFFSET_ARB, &glsl->min_gather_offset);
         get(GL_MAX_PROGRAM_TEXTURE_GATHER_OFFSET_ARB, &glsl->max_gather_offset);
     }
@@ -156,20 +176,25 @@ pl_gpu pl_gpu_create_gl(pl_log log, pl_opengl pl_gl, const struct pl_opengl_para
     struct pl_gpu_limits *limits = &gpu->limits;
     limits->thread_safe = params->make_current;
     limits->callbacks = gl_test_ext(gpu, "GL_ARB_sync", 32, 30);
-    if (gl_test_ext(gpu, "GL_ARB_pixel_buffer_object", 31, 0))
-        limits->max_buf_size = SIZE_MAX; // no restriction imposed by GL
-    if (gl_test_ext(gpu, "GL_ARB_uniform_buffer_object", 31, 0))
-        get(GL_MAX_UNIFORM_BLOCK_SIZE, &limits->max_ubo_size);
-    if (gl_test_ext(gpu, "GL_ARB_shader_storage_buffer_object", 43, 0))
-        get(GL_MAX_SHADER_STORAGE_BLOCK_SIZE, &limits->max_ssbo_size);
-    limits->max_vbo_size = limits->max_buf_size; // No additional restrictions
-    if (gl_test_ext(gpu, "GL_ARB_buffer_storage", 44, 0)) {
-        const char *vendor = (char *) gl->GetString(GL_VENDOR);
-        limits->max_mapped_size = limits->max_buf_size;
-        limits->host_cached = strcmp(vendor, "AMD") == 0 ||
-                              strcmp(vendor, "NVIDIA Corporation") == 0;
-    }
     limits->align_vertex_stride = 1;
+    if (gl_test_ext(gpu, "GL_ARB_pixel_buffer_object", 31, 0)) {
+        limits->max_buf_size = SIZE_MAX; // no restriction imposed by GL
+        if (gl_test_ext(gpu, "GL_ARB_uniform_buffer_object", 31, 0))
+            get(GL_MAX_UNIFORM_BLOCK_SIZE, &limits->max_ubo_size);
+        if (gl_test_ext(gpu, "GL_ARB_shader_storage_buffer_object", 43, 0) &&
+            gpu->glsl.version >= 140)
+        {
+            get(GL_MAX_SHADER_STORAGE_BLOCK_SIZE, &limits->max_ssbo_size);
+        }
+        limits->max_vbo_size = limits->max_buf_size; // No additional restrictions
+        if (gl_test_ext(gpu, "GL_ARB_buffer_storage", 44, 0)) {
+            const char *vendor = (char *) gl->GetString(GL_VENDOR);
+            limits->max_mapped_size = limits->max_buf_size;
+            limits->max_mapped_vram = limits->max_buf_size;
+            limits->host_cached = strcmp(vendor, "AMD") == 0 ||
+                                  strcmp(vendor, "NVIDIA Corporation") == 0;
+        }
+    }
 
     get(GL_MAX_TEXTURE_SIZE, &limits->max_tex_2d_dim);
     if (gl_test_ext(gpu, "GL_EXT_texture3D", 21, 30))
@@ -212,15 +237,12 @@ pl_gpu pl_gpu_create_gl(pl_log log, pl_opengl pl_gl, const struct pl_opengl_para
     }
 
     // Cache some internal capability checks
-    p->has_stride = gl_test_ext(gpu, "GL_EXT_unpack_subimage", 11, 30);
-    p->has_unpack_image_height = p->gl_ver >= 12 || p->gles_ver >= 30;
-    p->has_vao = gl_test_ext(gpu, "GL_ARB_vertex_array_object", 30, 0);
+    p->has_vao = gl_test_ext(gpu, "GL_ARB_vertex_array_object", 30, 30);
     p->has_invalidate_fb = gl_test_ext(gpu, "GL_ARB_invalidate_subdata", 43, 30);
     p->has_invalidate_tex = gl_test_ext(gpu, "GL_ARB_invalidate_subdata", 43, 0);
-    p->has_queries = gl_test_ext(gpu, "GL_ARB_timer_query", 33, 0);
-    p->has_fbos = gl_test_ext(gpu, "GL_ARB_framebuffer_object", 30, 20);
-    p->has_storage = gl_test_ext(gpu, "GL_ARB_shader_image_load_store", 42, 0);
-    p->has_readback = p->has_fbos;
+    p->has_queries = gl_test_ext(gpu, "GL_ARB_timer_query", 30, 30);
+    p->has_storage = gl_test_ext(gpu, "GL_ARB_shader_image_load_store", 42, 31);
+    p->has_readback = true;
 
     if (p->has_readback && p->gles_ver) {
         GLuint fbo = 0, tex = 0;
@@ -250,11 +272,9 @@ pl_gpu pl_gpu_create_gl(pl_log log, pl_opengl pl_gl, const struct pl_opengl_para
 
     // We simply don't know, so make up some values
     limits->align_tex_xfer_offset = 32;
-    limits->align_tex_xfer_pitch = 1;
+    limits->align_tex_xfer_pitch = 4;
     limits->fragment_queues = 1;
     limits->compute_queues = glsl->compute ? 1 : 0;
-    if (gl_test_ext(gpu, "GL_EXT_unpack_subimage", 11, 30))
-        limits->align_tex_xfer_pitch = 4;
 
     if (!gl_check_err(gpu, "pl_gpu_create_gl")) {
         PL_WARN(gpu, "Encountered errors while detecting GPU capabilities... "
@@ -532,17 +552,25 @@ static uint64_t gl_timer_query(pl_gpu gpu, pl_timer timer)
     if (timer->index_read == timer->index_write)
         return 0; // no more unprocessed results
 
+    struct pl_gl *p = PL_PRIV(gpu);
+
     const gl_funcs *gl = gl_funcs_get(gpu);
     if (!MAKE_CURRENT())
         return 0;
 
     uint64_t res = 0;
     GLuint query = timer->query[timer->index_read];
-    int avail = 0;
-    gl->GetQueryObjectiv(query, GL_QUERY_RESULT_AVAILABLE, &avail);
+    GLuint avail = 0;
+    gl->GetQueryObjectuiv(query, GL_QUERY_RESULT_AVAILABLE, &avail);
     if (!avail)
         goto done;
-    gl->GetQueryObjectui64v(query, GL_QUERY_RESULT, &res);
+    if (p->gles_ver || p->gl_ver < 33) {
+        GLuint tmp = 0;
+        gl->GetQueryObjectuiv(query, GL_QUERY_RESULT, &tmp);
+        res = tmp;
+    } else {
+        gl->GetQueryObjectui64v(query, GL_QUERY_RESULT, &res);
+    }
 
     timer->index_read = (timer->index_read + 1) % QUERY_OBJECT_NUM;
     // fall through
