@@ -427,10 +427,10 @@ ident_t sh_const(pl_shader sh, struct pl_shader_const sc)
         GLSLH("const int "$" = %d; \n", id, *(int *) sc.data);
         return id;
     case PL_VAR_UINT:
-        GLSLH("const uint "$" = %uu; \n", id, *(unsigned int *) sc.data);
+        GLSLH("const uint "$" = uint(%u); \n", id, *(unsigned int *) sc.data);
         return id;
     case PL_VAR_FLOAT:
-        GLSLH("const float "$" = %f; \n", id, *(float *) sc.data);
+        GLSLH("const float "$" = float(%f); \n", id, *(float *) sc.data);
         return id;
     case PL_VAR_INVALID:
     case PL_VAR_TYPE_COUNT:
@@ -506,7 +506,7 @@ ident_t sh_attr_vec2(pl_shader sh, const char *name, const pl_rect2df *rc)
     return sh_attr(sh, (struct pl_shader_va) {
         .attr = {
             .name     = name,
-            .fmt      = pl_find_vertex_fmt(gpu, PL_FMT_FLOAT, 2),
+            .fmt      = fmt,
         },
         .data = { verts[0], verts[1], verts[2], verts[3] },
     });
