@@ -16,40 +16,22 @@
  */
 
 #include <math.h>
+
+#include "cache.h"
+#include "colorspace.h"
 #include "shaders.h"
 
 #include <libplacebo/shaders/colorspace.h>
 
-// Common constants for SMPTE ST.2084 (PQ)
-static const float PQ_M1 = 2610./4096 * 1./4,
-                   PQ_M2 = 2523./4096 * 128,
-                   PQ_C1 = 3424./4096,
-                   PQ_C2 = 2413./4096 * 32,
-                   PQ_C3 = 2392./4096 * 32;
-
-// Common constants for ARIB STD-B67 (HLG)
-static const float HLG_A = 0.17883277,
-                   HLG_B = 0.28466892,
-                   HLG_C = 0.55991073,
-                   HLG_REF = 1000.0 / PL_COLOR_SDR_WHITE;
-
-// Common constants for Panasonic V-Log
-static const float VLOG_B = 0.00873,
-                   VLOG_C = 0.241514,
-                   VLOG_D = 0.598206;
-
-// Common constants for Sony S-Log
-static const float SLOG_A = 0.432699,
-                   SLOG_B = 0.037584,
-                   SLOG_C = 0.616596 + 0.03,
-                   SLOG_P = 3.538813,
-                   SLOG_Q = 0.030001,
-                   SLOG_K2 = 155.0 / 219.0;
-
 void pl_shader_set_alpha(pl_shader sh, struct pl_color_repr *repr,
                          enum pl_alpha_mode mode)
 {
-    if (repr->alpha == PL_ALPHA_PREMULTIPLIED && mode == PL_ALPHA_INDEPENDENT) {
+    bool src_has_alpha = repr->alpha == PL_ALPHA_INDEPENDENT ||
+                         repr->alpha == PL_ALPHA_PREMULTIPLIED;
+    bool dst_not_premul = mode == PL_ALPHA_INDEPENDENT ||
+                          mode == PL_ALPHA_NONE;
+
+    if (repr->alpha == PL_ALPHA_PREMULTIPLIED && dst_not_premul) {
         GLSL("if (color.a > 1e-6)               \n"
              "    color.rgb /= vec3(color.a);   \n");
         repr->alpha = PL_ALPHA_INDEPENDENT;
@@ -59,6 +41,11 @@ void pl_shader_set_alpha(pl_shader sh, struct pl_color_repr *repr,
         GLSL("color.rgb *= vec3(color.a); \n");
         repr->alpha = PL_ALPHA_PREMULTIPLIED;
     }
+
+    if (src_has_alpha && mode == PL_ALPHA_NONE) {
+        GLSL("color.a = 1.0; \n");
+        repr->alpha = PL_ALPHA_NONE;
+    }
 }
 
 #ifdef PL_HAVE_DOVI
@@ -601,10 +588,10 @@ void pl_shader_encode_color(pl_shader sh, const struct pl_color_repr *repr)
     GLSL("}\n");
 }
 
-static ident_t sh_luma_coeffs(pl_shader sh, const struct pl_raw_primaries *prim)
+static ident_t sh_luma_coeffs(pl_shader sh, const struct pl_color_space *csp)
 {
     pl_matrix3x3 rgb2xyz;
-    rgb2xyz = pl_get_rgb2xyz_matrix(prim);
+    rgb2xyz = pl_get_rgb2xyz_matrix(pl_raw_primaries_get(csp->primaries));
 
     // FIXME: Cannot use `const vec3` due to glslang bug #2025
     ident_t coeffs = sh_fresh(sh, "luma_coeffs");
@@ -707,9 +694,7 @@ void pl_shader_linearize(pl_shader sh, const struct pl_color_space *csp)
         // OOTF
         GLSL("color.rgb *= 1.0 / 12.0;                                      \n"
              "color.rgb *= "$" * pow(max(dot("$", color.rgb), 0.0), "$");   \n",
-             SH_FLOAT(csp_max),
-             sh_luma_coeffs(sh, pl_raw_primaries_get(csp->primaries)),
-             SH_FLOAT(y - 1));
+             SH_FLOAT(csp_max), sh_luma_coeffs(sh, csp), SH_FLOAT(y - 1));
         return;
     }
     case PL_COLOR_TRC_V_LOG:
@@ -763,33 +748,13 @@ void pl_shader_delinearize(pl_shader sh, const struct pl_color_space *csp)
     ));
 
     GLSL("// pl_shader_delinearize \n");
-    switch (csp->transfer) {
-    case PL_COLOR_TRC_UNKNOWN:
-    case PL_COLOR_TRC_SRGB:
-    case PL_COLOR_TRC_LINEAR:
-    case PL_COLOR_TRC_GAMMA18:
-    case PL_COLOR_TRC_GAMMA20:
-    case PL_COLOR_TRC_GAMMA22:
-    case PL_COLOR_TRC_GAMMA24:
-    case PL_COLOR_TRC_GAMMA26:
-    case PL_COLOR_TRC_GAMMA28:
-    case PL_COLOR_TRC_PRO_PHOTO:
-    case PL_COLOR_TRC_ST428: ;
-        if (csp_max != 1 || csp_min != 0) {
-            GLSL("color.rgb = "$" * color.rgb + vec3("$"); \n",
-                 SH_FLOAT(1 / (csp_max - csp_min)),
-                 SH_FLOAT(-csp_min / (csp_max - csp_min)));
-        }
-        break;
-    case PL_COLOR_TRC_BT_1886:
-    case PL_COLOR_TRC_PQ:
-    case PL_COLOR_TRC_HLG:
-    case PL_COLOR_TRC_V_LOG:
-    case PL_COLOR_TRC_S_LOG1:
-    case PL_COLOR_TRC_S_LOG2:
-        break; // scene-referred or absolute scale
-    case PL_COLOR_TRC_COUNT:
-        pl_unreachable();
+    if (pl_color_space_is_black_scaled(csp) &&
+        csp->transfer != PL_COLOR_TRC_HLG &&
+        (csp_max != 1 || csp_min != 0))
+    {
+        GLSL("color.rgb = "$" * color.rgb + vec3("$"); \n",
+             SH_FLOAT(1 / (csp_max - csp_min)),
+             SH_FLOAT(-csp_min / (csp_max - csp_min)));
     }
 
     GLSL("color.rgb = max(color.rgb, 0.0); \n");
@@ -851,9 +816,7 @@ void pl_shader_delinearize(pl_shader sh, const struct pl_color_space *csp)
         // OOTF^-1
         GLSL("color.rgb *= 1.0 / "$";                                       \n"
              "color.rgb *= 12.0 * max(1e-6, pow(dot("$", color.rgb), "$")); \n",
-             SH_FLOAT(csp_max),
-             sh_luma_coeffs(sh, pl_raw_primaries_get(csp->primaries)),
-             SH_FLOAT((1 - y) / y));
+             SH_FLOAT(csp_max), sh_luma_coeffs(sh, csp), SH_FLOAT((1 - y) / y));
         // OETF
         GLSL("color.rgb = mix(vec3(0.5) * sqrt(color.rgb),                      \n"
              "                vec3(%f) * log(color.rgb - vec3(%f)) + vec3(%f),  \n"
@@ -897,8 +860,8 @@ void pl_shader_sigmoidize(pl_shader sh, const struct pl_sigmoid_params *params)
         return;
 
     params = PL_DEF(params, &pl_sigmoid_default_params);
-    float center = PL_DEF(params->center, 0.75);
-    float slope  = PL_DEF(params->slope, 6.5);
+    float center = PL_DEF(params->center, pl_sigmoid_default_params.center);
+    float slope  = PL_DEF(params->slope, pl_sigmoid_default_params.slope);
 
     // This function needs to go through (0,0) and (1,1), so we compute the
     // values at 1 and 0, and then scale/shift them, respectively.
@@ -921,8 +884,8 @@ void pl_shader_unsigmoidize(pl_shader sh, const struct pl_sigmoid_params *params
 
     // See: pl_shader_sigmoidize
     params = PL_DEF(params, &pl_sigmoid_default_params);
-    float center = PL_DEF(params->center, 0.75);
-    float slope  = PL_DEF(params->slope, 6.5);
+    float center = PL_DEF(params->center, pl_sigmoid_default_params.center);
+    float slope  = PL_DEF(params->slope, pl_sigmoid_default_params.slope);
     float offset = 1.0 / (1 + expf(slope * center));
     float scale  = 1.0 / (1 + expf(slope * (center - 1))) - offset;
 
@@ -945,7 +908,6 @@ static bool peak_detect_params_eq(const struct pl_peak_detect_params *a,
     return a->smoothing_period     == b->smoothing_period     &&
            a->scene_threshold_low  == b->scene_threshold_low  &&
            a->scene_threshold_high == b->scene_threshold_high &&
-           a->minimum_peak         == b->minimum_peak         &&
            a->percentile           == b->percentile;
     // don't compare `allow_delayed` because it doesn't change measurement
 }
@@ -979,6 +941,7 @@ pl_static_assert(PQ_BITS >= HIST_BITS);
 
 struct peak_buf_data {
     unsigned frame_wg_count[SLICES]; // number of work groups processed
+    unsigned frame_wg_active[SLICES];// number of active (nonzero) work groups
     unsigned frame_sum_pq[SLICES];   // sum of PQ Y values over all WGs (PQ_BITS)
     unsigned frame_max_pq[SLICES];   // maximum PQ Y value among these WGs (PQ_BITS)
     unsigned frame_hist[SLICES][HIST_BINS]; // always allocated, conditionally used
@@ -1001,6 +964,7 @@ static const struct pl_buffer_var peak_buf_vars[] = {
     },                                                                          \
 }
     VAR(frame_wg_count),
+    VAR(frame_wg_active),
     VAR(frame_sum_pq),
     VAR(frame_max_pq),
     VAR(frame_hist),
@@ -1016,7 +980,6 @@ struct sh_color_map_obj {
 
     // Gamut map state
     struct {
-        struct pl_gamut_map_params params;
         pl_shader_obj lut;
     } gamut;
 
@@ -1030,6 +993,19 @@ struct sh_color_map_obj {
     } peak;
 };
 
+// Excluding size, since this is checked by sh_lut
+static uint64_t gamut_map_signature(const struct pl_gamut_map_params *par)
+{
+    uint64_t sig = CACHE_KEY_GAMUT_LUT;
+    pl_hash_merge(&sig, pl_str0_hash(par->function->name));
+    pl_hash_merge(&sig, pl_var_hash(par->input_gamut));
+    pl_hash_merge(&sig, pl_var_hash(par->output_gamut));
+    pl_hash_merge(&sig, pl_var_hash(par->min_luma));
+    pl_hash_merge(&sig, pl_var_hash(par->max_luma));
+    pl_hash_merge(&sig, pl_var_hash(par->constants));
+    return sig;
+}
+
 static void sh_color_map_uninit(pl_gpu gpu, void *ptr)
 {
     struct sh_color_map_obj *obj = ptr;
@@ -1042,8 +1018,9 @@ static void sh_color_map_uninit(pl_gpu gpu, void *ptr)
 
 static inline float iir_coeff(float rate)
 {
-    float a = 1.0 - cos(1.0 / rate);
-    return sqrt(a*a + 2*a) - a;
+    if (!rate)
+        return 1.0f;
+    return 1.0f - expf(-1.0f / rate);
 }
 
 static float measure_peak(const struct peak_buf_data *data, float percentile)
@@ -1120,30 +1097,35 @@ static void update_peak_buf(pl_gpu gpu, struct sh_color_map_obj *obj, bool force
         pl_buf_destroy(gpu, &obj->peak.buf);
     } else {
         // No data read? Possibly this peak obj has not been executed yet
-        if (params->allow_delayed) {
-            PL_TRACE(gpu, "Peak detection buffer seems empty, ignoring..");
-        } else if (ok) {
+        if (!ok) {
+            PL_ERR(gpu, "Failed reading peak detection buffer!");
+        } else if (params->allow_delayed) {
+            PL_TRACE(gpu, "Peak detection buffer not yet ready, ignoring..");
+        } else {
             PL_WARN(gpu, "Peak detection usage error: attempted detecting peak "
                     "and using detected peak in the same shader program, "
                     "but `params->allow_delayed` is false! Ignoring, but "
                     "expect incorrect output.");
-        } else {
-            PL_ERR(gpu, "Failed reading peak detection buffer!");
         }
-        if (force)
+        if (force || !ok)
             pl_buf_destroy(gpu, &obj->peak.buf);
         return;
     }
 
-    uint64_t frame_sum_pq = 0u, frame_wg_count = 0u;
+    uint64_t frame_sum_pq = 0u, frame_wg_count = 0u, frame_wg_active = 0u;
     for (int k = 0; k < SLICES; k++) {
-        frame_sum_pq   += data.frame_sum_pq[k];
-        frame_wg_count += data.frame_wg_count[k];
+        frame_sum_pq    += data.frame_sum_pq[k];
+        frame_wg_count  += data.frame_wg_count[k];
+        frame_wg_active += data.frame_wg_active[k];
+    }
+    float avg_pq, max_pq;
+    if (frame_wg_active) {
+        avg_pq = (float) frame_sum_pq / (frame_wg_active * PQ_MAX);
+        max_pq = measure_peak(&data, params->percentile);
+    } else {
+        // Solid black frame
+        avg_pq = max_pq = PL_COLOR_HDR_BLACK;
     }
-    float avg_pq = (float) frame_sum_pq / (frame_wg_count * PQ_MAX);
-    float max_pq = measure_peak(&data, params->percentile);
-    const float min_peak = PL_DEF(params->minimum_peak, 1.0f);
-    max_pq = fmaxf(max_pq, pl_hdr_rescale(PL_HDR_NORM, PL_HDR_PQ, min_peak));
 
     if (!obj->peak.avg_pq) {
         // Set the initial value accordingly if it contains no data
@@ -1159,7 +1141,7 @@ static void update_peak_buf(pl_gpu gpu, struct sh_color_map_obj *obj, bool force
     }
 
     // Use an IIR low-pass filter to smooth out the detected values
-    const float coeff = iir_coeff(PL_DEF(params->smoothing_period, 100.0f));
+    const float coeff = iir_coeff(params->smoothing_period);
     obj->peak.avg_pq += coeff * (avg_pq - obj->peak.avg_pq);
     obj->peak.max_pq += coeff * (max_pq - obj->peak.max_pq);
 
@@ -1168,7 +1150,8 @@ static void update_peak_buf(pl_gpu gpu, struct sh_color_map_obj *obj, bool force
         const float log10_pq = 1e-2f; // experimentally determined approximate
         const float thresh_low = params->scene_threshold_low * log10_pq;
         const float thresh_high = params->scene_threshold_high * log10_pq;
-        const float delta = fabsf(avg_pq - obj->peak.avg_pq);
+        const float bias = (float) frame_wg_active / frame_wg_count;
+        const float delta = bias * fabsf(avg_pq - obj->peak.avg_pq);
         const float mix_coeff = pl_smoothstep(thresh_low, thresh_high, delta);
         obj->peak.avg_pq = PL_MIX(obj->peak.avg_pq, avg_pq, mix_coeff);
         obj->peak.max_pq = PL_MIX(obj->peak.max_pq, max_pq, mix_coeff);
@@ -1192,7 +1175,7 @@ bool pl_shader_detect_peak(pl_shader sh, struct pl_color_space csp,
     }
 
     const bool use_histogram = params->percentile > 0 && params->percentile < 100;
-    size_t shmem_req = 2 * sizeof(uint32_t);
+    size_t shmem_req = 3 * sizeof(uint32_t);
     if (use_histogram)
         shmem_req += sizeof(uint32_t[HIST_BINS]);
 
@@ -1264,102 +1247,113 @@ retry_ssbo:
         .num_buffer_vars = PL_ARRAY_SIZE(peak_buf_vars),
     });
 
-    sh_describe(sh, "peak detection");
-    GLSL("// pl_shader_detect_peak                                      \n"
-         "{                                                             \n"
-         "const uint wg_size = gl_WorkGroupSize.x * gl_WorkGroupSize.y; \n"
-         "uint wg_idx = gl_WorkGroupID.y * gl_NumWorkGroups.x +         \n"
-         "              gl_WorkGroupID.x;                               \n"
-         "uint slice = wg_idx %% %du;                                   \n"
-         "vec4 color_orig = color;                                      \n",
-         SLICES);
-
     // For performance, we want to do as few atomic operations on global
     // memory as possible, so use an atomic in shmem for the work group.
-    ident_t wg_sum = sh_fresh(sh, "wg_sum"),
-            wg_max = sh_fresh(sh, "wg_max"),
-            wg_hist = NULL_IDENT;
-    GLSLH("shared uint "$", "$"; \n", wg_sum, wg_max);
+    ident_t wg_sum   = sh_fresh(sh, "wg_sum"),
+            wg_max   = sh_fresh(sh, "wg_max"),
+            wg_black = sh_fresh(sh, "wg_black"),
+            wg_hist  = NULL_IDENT;
+    GLSLH("shared uint "$", "$", "$"; \n", wg_sum, wg_max, wg_black);
     if (use_histogram) {
         wg_hist = sh_fresh(sh, "wg_hist");
         GLSLH("shared uint "$"[%u]; \n", wg_hist, HIST_BINS);
-        GLSL("for (uint i = gl_LocalInvocationIndex; i < %du; i += wg_size) \n"
-             "    "$"[i] = 0u;                                              \n",
-             HIST_BINS, wg_hist);
     }
-    GLSL($" = 0u; "$" = 0u; \n"
-         "barrier();        \n",
-         wg_sum, wg_max);
+
+    sh_describe(sh, "peak detection");
+#pragma GLSL /* pl_shader_detect_peak */                                        \
+    {                                                                           \
+    const uint wg_size = gl_WorkGroupSize.x * gl_WorkGroupSize.y;               \
+    const uint wg_idx = gl_WorkGroupID.y * gl_NumWorkGroups.x + gl_WorkGroupID.x;\
+    const uint local_idx = gl_LocalInvocationIndex;                             \
+    const uint slice = wg_idx % ${const uint: SLICES};                          \
+    const uint hist_base = slice * ${const uint: HIST_BINS};                    \
+    const vec4 color_orig = color;                                              \
+    $wg_sum = $wg_max = $wg_black = 0u;                                         \
+    @if (use_histogram) {                                                       \
+        for (uint i = local_idx; i < ${const uint: HIST_BINS}; i += wg_size)    \
+            $wg_hist[i] = 0u;                                                   \
+    @}                                                                          \
+    barrier();
 
     // Decode color into linear light representation
     pl_color_space_infer(&csp);
     pl_shader_linearize(sh, &csp);
 
-    // Measure luminance as N-bit PQ
-    GLSL("float luma = dot("$", color.rgb);             \n"
-         "luma *= %f;                                   \n"
-         "luma = pow(clamp(luma, 0.0, 1.0), %f);        \n"
-         "luma = (%f + %f * luma) / (1.0 + %f * luma);  \n"
-         "luma = pow(luma, %f);                         \n"
-         "uint y_pq = uint(%d.0 * luma);                \n",
-         sh_luma_coeffs(sh, &csp.hdr.prim),
-         PL_COLOR_SDR_WHITE / 10000.0,
-         PQ_M1, PQ_C1, PQ_C2, PQ_C3, PQ_M2,
-         PQ_MAX);
-
-    // Update the work group's shared atomics
     bool has_subgroups = sh_glsl(sh).subgroup_size > 0;
-    if (use_histogram) {
-        GLSL("int bin = (int(y_pq) >> %d) - %d; \n"
-             "bin = clamp(bin, 0, %d);          \n",
-             PQ_BITS - HIST_BITS, HIST_BIAS,
-             HIST_BINS - 1);
-        if (has_subgroups) {
-            // Optimize for the very common case of identical histogram bins
-            GLSL("if (subgroupAllEqual(bin)) {                  \n"
-                 "    if (subgroupElect())                      \n"
-                 "        atomicAdd("$"[bin], gl_SubgroupSize); \n"
-                 "} else {                                      \n"
-                 "    atomicAdd("$"[bin], 1u);                  \n"
-                 "}                                             \n",
-                 wg_hist, wg_hist);
-        } else {
-            GLSL("atomicAdd("$"[bin], 1u); \n", wg_hist);
-        }
+    const float cutoff = fmaxf(params->black_cutoff, 0.0f) * 1e-2f;
+#pragma GLSL /* Measure luminance as N-bit PQ */                                \
+    float luma = dot(${sh_luma_coeffs(sh, &csp)}, color.rgb);                   \
+    luma *= ${const float: PL_COLOR_SDR_WHITE / 10000.0};                       \
+    luma = pow(clamp(luma, 0.0, 1.0), ${const float: PQ_M1});                   \
+    luma = (${const float: PQ_C1} + ${const float: PQ_C2} * luma) /             \
+           (1.0 + ${const float: PQ_C3} * luma);                                \
+    luma = pow(luma, ${const float: PQ_M2});                                    \
+    @if (cutoff)                                                                \
+        luma *= smoothstep(0.0, ${float: cutoff}, luma);                        \
+    uint y_pq = uint(${const float: PQ_MAX} * luma);                            \
+                                                                                \
+    /* Update the work group's shared atomics */                                \
+    @if (use_histogram) {                                                       \
+        int bin = int(y_pq) >> ${const int: PQ_BITS - HIST_BITS};               \
+        bin -= ${const int: HIST_BIAS};                                         \
+        bin = clamp(bin, 0, ${const int: HIST_BINS - 1});                       \
+        @if (has_subgroups) {                                                   \
+            /* Optimize for the very common case of identical histogram bins */ \
+            if (subgroupAllEqual(bin)) {                                        \
+                if (subgroupElect())                                            \
+                    atomicAdd($wg_hist[bin], gl_SubgroupSize);                  \
+            } else {                                                            \
+                atomicAdd($wg_hist[bin], 1u);                                   \
+            }                                                                   \
+        @} else {                                                               \
+            atomicAdd($wg_hist[bin], 1u);                                       \
+        @}                                                                      \
+    @}                                                                          \
+                                                                                \
+    @if (has_subgroups) {                                                       \
+        uint group_sum = subgroupAdd(y_pq);                                     \
+        uint group_max = subgroupMax(y_pq);                                     \
+        @if (cutoff)                                                            \
+            uvec4 b = subgroupBallot(y_pq == 0u);                               \
+        if (subgroupElect()) {                                                  \
+            atomicAdd($wg_sum, group_sum);                                      \
+            atomicMax($wg_max, group_max);                                      \
+            @if (cutoff)                                                        \
+                atomicAdd($wg_black, subgroupBallotBitCount(b));                \
+        }                                                                       \
+    @} else {                                                                   \
+        atomicAdd($wg_sum, y_pq);                                               \
+        atomicMax($wg_max, y_pq);                                               \
+        @if (cutoff) {                                                          \
+            if (y_pq == 0u)                                                     \
+                atomicAdd($wg_black, 1u);                                       \
+        @}                                                                      \
+    @}                                                                          \
+    barrier();                                                                  \
+                                                                                \
+    @if (use_histogram) {                                                       \
+        @if (cutoff) {                                                          \
+            if (gl_LocalInvocationIndex == 0u)                                  \
+                $wg_hist[0] -= $wg_black;                                       \
+        @}                                                                      \
+        /* Update the histogram with a cooperative loop */                      \
+        for (uint i = local_idx; i < ${const uint: HIST_BINS}; i += wg_size)    \
+            atomicAdd(frame_hist[hist_base + i], $wg_hist[i]);                  \
+    @}                                                                          \
+                                                                                \
+    /* Have one thread per work group update the global atomics */              \
+    if (gl_LocalInvocationIndex == 0u) {                                        \
+        uint num = wg_size - $wg_black;                                         \
+        atomicAdd(frame_wg_count[slice], 1u);                                   \
+        atomicAdd(frame_wg_active[slice], min(num, 1u));                        \
+        if (num > 0u) {                                                         \
+            atomicAdd(frame_sum_pq[slice], $wg_sum / num);                      \
+            atomicMax(frame_max_pq[slice], $wg_max);                            \
+        }                                                                       \
+    }                                                                           \
+    color = color_orig;                                                         \
     }
 
-    if (has_subgroups) {
-        GLSL("uint group_sum = subgroupAdd(y_pq);   \n"
-             "uint group_max = subgroupMax(y_pq);   \n"
-             "if (subgroupElect()) {                \n"
-             "    atomicAdd("$", group_sum);        \n"
-             "    atomicMax("$", group_max);        \n"
-             "}                                     \n"
-             "barrier();                            \n",
-             wg_sum, wg_max);
-    } else {
-        GLSL("atomicAdd("$", y_pq); \n"
-             "atomicMax("$", y_pq); \n"
-             "barrier();            \n",
-             wg_sum, wg_max);
-    }
-
-    if (use_histogram) {
-        GLSL("for (uint i = gl_LocalInvocationIndex; i < %du; i += wg_size) \n"
-             "    atomicAdd(frame_hist[slice * %du + i], "$"[i]);           \n",
-             HIST_BINS, HIST_BINS, wg_hist);
-    }
-
-    // Have one thread per work group update the global atomics
-    GLSL("if (gl_LocalInvocationIndex == 0u) {              \n"
-         "    atomicAdd(frame_wg_count[slice], 1u);         \n"
-         "    atomicAdd(frame_sum_pq[slice], "$" / wg_size);\n"
-         "    atomicMax(frame_max_pq[slice], "$");          \n"
-         "}                                                 \n"
-         "color = color_orig;                               \n"
-         "}                                                 \n",
-          wg_sum, wg_max);
-
     return true;
 }
 
@@ -1379,19 +1373,6 @@ bool pl_get_detected_hdr_metadata(const pl_shader_obj state,
     return true;
 }
 
-bool pl_get_detected_peak(const pl_shader_obj state,
-                          float *out_peak, float *out_avg)
-{
-    struct pl_hdr_metadata data;
-    if (!pl_get_detected_hdr_metadata(state, &data))
-        return false;
-
-    // Preserves old behavior
-    *out_peak = pl_hdr_rescale(PL_HDR_PQ, PL_HDR_NORM, data.max_pq_y);
-    *out_avg  = pl_hdr_rescale(PL_HDR_PQ, PL_HDR_NORM, data.avg_pq_y);
-    return true;
-}
-
 void pl_reset_detected_peak(pl_shader_obj state)
 {
     if (!state || state->type != PL_SHADER_OBJ_COLOR_MAP)
@@ -1547,43 +1528,48 @@ static void visualize_gamut_map(pl_shader sh, pl_rect2df rc,
          // Check against src/dst gamut boundaries
          "vec3 rgbsrc = "$" * lms;                              \n"
          "vec3 rgbdst = "$" * lms;                              \n"
-         "bool ingamut_src, ingamut_dst;                        \n"
-         "ingamut_src = all(lessThan(rgbsrc, vec3(rgbmax))) &&  \n"
+         "bool insrc, indst;                                    \n"
+         "insrc = all(lessThan(rgbsrc, vec3(rgbmax))) &&        \n"
          "              all(greaterThan(rgbsrc, vec3(rgbmin))); \n"
-         "ingamut_dst = all(lessThan(rgbdst, vec3(rgbmax))) &&  \n"
+         "indst = all(lessThan(rgbdst, vec3(rgbmax))) &&        \n"
          "              all(greaterThan(rgbdst, vec3(rgbmin))); \n"
          // Sample from gamut mapping 3DLUT
          "idx.x = (ipt.x - pqmin) / (pqmax - pqmin);            \n"
          "idx.y = 2.0 * length(ipt.yz);                         \n"
          "idx.z = %f * atan(ipt.z, ipt.y) + 0.5;                \n"
-         "bool inlut = all(lessThanEqual(idx, vec3(1.0))) &&    \n"
-         "             all(greaterThanEqual(idx, vec3(0.0)));   \n"
          "vec3 mapped = "$"(idx).xyz;                           \n"
+         "mapped.yz -= vec2(32768.0/65535.0);                   \n"
          "float mappedhue = atan(mapped.z, mapped.y);           \n"
+         "float mappedchroma = length(mapped.yz);               \n"
          "ipt = mapped;                                         \n"
          // Visualize gamuts
-         "if (!ingamut_src && !ingamut_dst) {                   \n"
+         "if (!insrc && !indst) {                               \n"
          "    ipt = orig;                                       \n"
-         "} else if (ingamut_src && !ingamut_dst) {             \n"
+         "} else if (insrc && !indst) {                         \n"
          "    ipt.x -= 0.1;                                     \n"
-         "} else if (ingamut_dst && !ingamut_src) {             \n"
+         "} else if (indst && !insrc) {                         \n"
          "    ipt.x += 0.1;                                     \n"
          "}                                                     \n"
          // Visualize iso-luminance and iso-hue lines
          "vec3 line;                                            \n"
-         "if (inlut && fract(50.0 * mapped.x) < 1e-1) {         \n"
+         "if (insrc && fract(50.0 * mapped.x) < 1e-1) {         \n"
          "    float k = smoothstep(0.1, 0.0, abs(sin(theta)));  \n"
          "    line.x = mix(mapped.x, 0.3, 0.5);                 \n"
          "    line.yz = sqrt(length(mapped.yz)) *               \n"
          "              normalize(mapped.yz);                   \n"
          "    ipt = mix(ipt, line, k);                          \n"
          "}                                                     \n"
-         "if (inlut && fract(10.0 * (mappedhue - hue)) < 1e-1) {\n"
+         "if (insrc && fract(10.0 * (mappedhue - hue)) < 1e-1) {\n"
          "    float k = smoothstep(0.3, 0.0, abs(cos(theta)));  \n"
          "    line.x = mapped.x - 0.05;                         \n"
          "    line.yz = 1.2 * mapped.yz;                        \n"
          "    ipt = mix(ipt, line, k);                          \n"
          "}                                                     \n"
+         "if (insrc && fract(100.0 * mappedchroma) < 1e-1) {    \n"
+         "    line.x = mapped.x + 0.1;                          \n"
+         "    line.yz = 0.4 * mapped.yz;                        \n"
+         "    ipt = mix(ipt, line, 0.5);                        \n"
+         "}                                                     \n"
          "}                                                     \n",
          rect_pos(sh, rc),
          SH_FLOAT(params->min_luma), SH_FLOAT(params->max_luma),
@@ -1608,7 +1594,24 @@ static void fill_tone_lut(void *data, const struct sh_lut_params *params)
 static void fill_gamut_lut(void *data, const struct sh_lut_params *params)
 {
     const struct pl_gamut_map_params *lut_params = params->priv;
-    pl_gamut_map_generate(data, lut_params);
+    const int lut_size = params->width * params->height * params->depth;
+    void *tmp = pl_alloc(NULL, lut_size * sizeof(float) * lut_params->lut_stride);
+    pl_gamut_map_generate(tmp, lut_params);
+
+    // Convert to 16-bit unsigned integer for GPU texture
+    const float *in = tmp;
+    uint16_t *out = data;
+    pl_assert(lut_params->lut_stride == 3);
+    pl_assert(params->comps == 4);
+    for (int i = 0; i < lut_size; i++) {
+        out[0] = roundf(in[0] * UINT16_MAX);
+        out[1] = roundf(in[1] * UINT16_MAX + (UINT16_MAX >> 1));
+        out[2] = roundf(in[2] * UINT16_MAX + (UINT16_MAX >> 1));
+        in  += 3;
+        out += 4;
+    }
+
+    pl_free(tmp);
 }
 
 void pl_shader_color_map_ex(pl_shader sh, const struct pl_color_map_params *params,
@@ -1618,13 +1621,6 @@ void pl_shader_color_map_ex(pl_shader sh, const struct pl_color_map_params *para
         return;
 
     struct pl_color_space src = args->src, dst = args->dst;
-    pl_color_space_infer_map(&src, &dst);
-    if (pl_color_space_equal(&src, &dst)) {
-        if (args->prelinearized)
-            pl_shader_delinearize(sh, &dst);
-        return;
-    }
-
     struct sh_color_map_obj *obj = NULL;
     if (args->state) {
         pl_get_detected_hdr_metadata(*args->state, &src.hdr);
@@ -1634,17 +1630,25 @@ void pl_shader_color_map_ex(pl_shader sh, const struct pl_color_map_params *para
             return;
     }
 
+    pl_color_space_infer_map(&src, &dst);
+    if (pl_color_space_equal(&src, &dst)) {
+        if (args->prelinearized)
+            pl_shader_delinearize(sh, &dst);
+        return;
+    }
+
     params = PL_DEF(params, &pl_color_map_default_params);
     GLSL("// pl_shader_color_map \n"
          "{                      \n");
 
     struct pl_tone_map_params tone = {
-        .function = params->tone_mapping_function,
-        .param = params->tone_mapping_param,
-        .input_scaling = PL_HDR_PQ,
+        .function       = PL_DEF(params->tone_mapping_function, &pl_tone_map_clip),
+        .constants      = params->tone_constants,
+        .param          = params->tone_mapping_param,
+        .input_scaling  = PL_HDR_PQ,
         .output_scaling = PL_HDR_PQ,
-        .lut_size = PL_DEF(params->lut_size, pl_color_map_default_params.lut_size),
-        .hdr = src.hdr,
+        .lut_size       = PL_DEF(params->lut_size, pl_color_map_default_params.lut_size),
+        .hdr            = src.hdr,
     };
 
     pl_color_space_nominal_luma_ex(pl_nominal_luma_params(
@@ -1664,11 +1668,7 @@ void pl_shader_color_map_ex(pl_shader sh, const struct pl_color_map_params *para
         .out_max    = &tone.output_max,
     ));
 
-    if (!params->inverse_tone_mapping) {
-        // Never exceed the source unless requested, but still allow
-        // black point adaptation
-        tone.output_max = PL_MIN(tone.output_max, tone.input_max);
-    }
+    pl_tone_map_params_infer(&tone);
 
     // Round sufficiently similar values
     if (fabs(tone.input_max - tone.output_max) < 1e-6)
@@ -1676,16 +1676,22 @@ void pl_shader_color_map_ex(pl_shader sh, const struct pl_color_map_params *para
     if (fabs(tone.input_min - tone.output_min) < 1e-6)
         tone.output_min = tone.input_min;
 
-    pl_tone_map_params_infer(&tone);
+    if (!params->inverse_tone_mapping) {
+        // Never exceed the source unless requested, but still allow
+        // black point adaptation
+        tone.output_max = PL_MIN(tone.output_max, tone.input_max);
+    }
 
+    const int *lut3d_size_def = pl_color_map_default_params.lut3d_size;
     struct pl_gamut_map_params gamut = {
         .function        = PL_DEF(params->gamut_mapping, &pl_gamut_map_clip),
+        .constants       = params->gamut_constants,
         .input_gamut     = src.hdr.prim,
         .output_gamut    = dst.hdr.prim,
-        .lut_size_I      = PL_DEF(params->lut3d_size[0], 23),
-        .lut_size_C      = PL_DEF(params->lut3d_size[1], 32),
-        .lut_size_h      = PL_DEF(params->lut3d_size[2], 128),
-        .lut_stride      = 4,
+        .lut_size_I      = PL_DEF(params->lut3d_size[0], lut3d_size_def[0]),
+        .lut_size_C      = PL_DEF(params->lut3d_size[1], lut3d_size_def[1]),
+        .lut_size_h      = PL_DEF(params->lut3d_size[2], lut3d_size_def[2]),
+        .lut_stride      = 3,
     };
 
     float src_peak_static;
@@ -1704,6 +1710,14 @@ void pl_shader_color_map_ex(pl_shader sh, const struct pl_color_map_params *para
         .out_max    = &gamut.max_luma,
     ));
 
+    // Clip the gamut mapping output to the input gamut if disabled
+    if (!params->gamut_expansion && gamut.function->bidirectional) {
+        if (pl_primaries_compatible(&gamut.input_gamut, &gamut.output_gamut)) {
+            gamut.output_gamut = pl_primaries_clip(&gamut.output_gamut,
+                                                   &gamut.input_gamut);
+        }
+    }
+
     // Backwards compatibility with older API
     switch (params->gamut_mode) {
     case PL_GAMUT_CLIP:
@@ -1737,10 +1751,16 @@ void pl_shader_color_map_ex(pl_shader sh, const struct pl_color_map_params *para
     if (!args->state) {
         // No state object provided, forcibly disable advanced methods
         can_fast = true;
-        if (tone.function != &pl_tone_map_linear)
-            tone.function = &pl_tone_map_clip;
-        if (gamut.function != &pl_gamut_map_saturation)
-            gamut.function = &pl_gamut_map_clip;
+        if (tone.function != &pl_tone_map_clip)
+            tone.function = &pl_tone_map_linear;
+        if (gamut.function != &pl_gamut_map_clip)
+            gamut.function = &pl_gamut_map_saturation;
+    }
+
+    pl_fmt gamut_fmt = pl_find_fmt(SH_GPU(sh), PL_FMT_UNORM, 4, 16, 16, PL_FMT_CAP_LINEAR);
+    if (!gamut_fmt) {
+        gamut.function = &pl_gamut_map_saturation;
+        can_fast = true;
     }
 
     bool need_tone_map = !pl_tone_map_params_noop(&tone);
@@ -1754,7 +1774,7 @@ void pl_shader_color_map_ex(pl_shader sh, const struct pl_color_map_params *para
     ident_t lms2ipt = SH_MAT3(pl_ipt_lms2ipt);
     ident_t ipt2lms = SH_MAT3(pl_ipt_ipt2lms);
 
-    if (gamut.function == &pl_gamut_map_saturation && can_fast) {
+    if (need_gamut_map && gamut.function == &pl_gamut_map_saturation && can_fast) {
         const pl_matrix3x3 lms2src = pl_ipt_lms2rgb(&gamut.input_gamut);
         const pl_matrix3x3 dst2lms = pl_ipt_rgb2lms(&gamut.output_gamut);
         sh_describe(sh, "gamut map (saturation)");
@@ -1814,7 +1834,7 @@ void pl_shader_color_map_ex(pl_shader sh, const struct pl_color_map_params *para
 
         } else if (fun == &pl_tone_map_linear && can_fast) {
 
-            const float gain = PL_DEF(tone.param, 1.0f);
+            const float gain = tone.constants.exposure;
             const float scale = tone.input_max - tone.input_min;
 
             ident_t linfun = sh_fresh(sh, "linear_pq");
@@ -1898,18 +1918,23 @@ void pl_shader_color_map_ex(pl_shader sh, const struct pl_color_map_params *para
                  "float detail = highres - lowres;                  \n"
                  "float base = tone_map(highres);                   \n"
                  "float sharp = tone_map(lowres) + detail;          \n"
-                 "ipt.x = mix(base, sharp, "$");                    \n",
+                 "ipt.x = clamp(mix(base, sharp, "$"), "$", "$");   \n",
                  pos, pt, lowres,
                  lowres, lowres, lowres, lowres,
-                 SH_FLOAT(params->contrast_recovery));
+                 SH_FLOAT(params->contrast_recovery),
+                 SH_FLOAT(tone.output_min), SH_FLOAT_DYN(tone.output_max));
 
         } else {
 
             GLSL("ipt.x = tone_map(ipt.x); \n");
         }
 
-        // Avoid raising saturation excessively when changing brightness
-        GLSL("ipt.yz *= min(i_orig / ipt.x, ipt.x / i_orig); \n");
+        // Avoid raising saturation excessively when raising brightness, and
+        // also desaturate when reducing brightness greatly to account for the
+        // reduction in gamut volume.
+        GLSL("vec2 hull = vec2(i_orig, ipt.x);                  \n"
+             "hull = ((hull - 6.0) * hull + 9.0) * hull;        \n"
+             "ipt.yz *= min(i_orig / ipt.x, hull.y / hull.x);   \n");
     }
 
     if (need_gamut_map) {
@@ -1921,16 +1946,17 @@ void pl_shader_color_map_ex(pl_shader sh, const struct pl_color_map_params *para
             .object     = &obj->gamut.lut,
             .var_type   = PL_VAR_FLOAT,
             .lut_type   = SH_LUT_TEXTURE,
+            .fmt        = gamut_fmt,
             .method     = params->lut3d_tricubic ? SH_LUT_CUBIC : SH_LUT_LINEAR,
             .width      = gamut.lut_size_I,
             .height     = gamut.lut_size_C,
             .depth      = gamut.lut_size_h,
-            .comps      = gamut.lut_stride,
-            .update     = !pl_gamut_map_params_equal(&gamut, &obj->gamut.params),
+            .comps      = 4,
+            .signature  = gamut_map_signature(&gamut),
+            .cache      = SH_CACHE(sh),
             .fill       = fill_gamut_lut,
             .priv       = &gamut,
         ));
-        obj->gamut.params = gamut;
         if (!lut) {
             SH_FAIL(sh, "Failed generating gamut-mapping LUT!");
             return;
@@ -1942,7 +1968,8 @@ void pl_shader_color_map_ex(pl_shader sh, const struct pl_color_map_params *para
              "idx.x = "$" * ipt.x + "$";            \n"
              "idx.y = 2.0 * length(ipt.yz);         \n"
              "idx.z = %f * atan(ipt.z, ipt.y) + 0.5;\n"
-             "ipt = "$"(idx).xyz;                   \n",
+             "ipt = "$"(idx).xyz;                   \n"
+             "ipt.yz -= vec2(32768.0/65535.0);      \n",
              SH_FLOAT(1.0f / lut_range),
              SH_FLOAT(-gamut.min_luma / lut_range),
              0.5f / M_PI, lut);
