@@ -1,4 +1,7 @@
-#include "tests.h"
+#include "utils.h"
+
+#include <libplacebo/dither.h>
+#include <libplacebo/shaders/dithering.h>
 
 #define SHIFT 4
 #define SIZE (1 << SHIFT)
