@@ -19,7 +19,9 @@
 
 #define VK_NO_PROTOTYPES
 #define VK_ENABLE_BETA_EXTENSIONS // for VK_KHR_portability_subset
+#ifdef __APPLE__
 #define VK_USE_PLATFORM_METAL_EXT
+#endif
 
 #include "../common.h"
 #include "../log.h"
@@ -179,7 +181,6 @@ struct vk_ctx {
     PL_VK_FUN(DestroyFramebuffer);
     PL_VK_FUN(DestroyImage);
     PL_VK_FUN(DestroyImageView);
-    PL_VK_FUN(DestroyInstance);
     PL_VK_FUN(DestroyPipeline);
     PL_VK_FUN(DestroyPipelineCache);
     PL_VK_FUN(DestroyPipelineLayout);
