@@ -187,13 +187,6 @@ static const struct vk_ext vk_device_extensions[] = {
             PL_VK_DEV_FUN(QueueSubmit2KHR),
             {0}
         },
-    }, {
-        .name = VK_KHR_DYNAMIC_RENDERING_EXTENSION_NAME,
-        .funs = (const struct vk_fun[]) {
-            PL_VK_DEV_FUN(CmdBeginRenderingKHR),
-            PL_VK_DEV_FUN(CmdEndRenderingKHR),
-            {0}
-        },
     },
 };
 
@@ -221,7 +214,6 @@ const char * const pl_vulkan_recommended_extensions[] = {
     VK_EXT_FULL_SCREEN_EXCLUSIVE_EXTENSION_NAME,
 #endif
     VK_KHR_SYNCHRONIZATION_2_EXTENSION_NAME,
-    VK_KHR_DYNAMIC_RENDERING_EXTENSION_NAME,
 };
 
 const int pl_vulkan_num_recommended_extensions =
@@ -239,6 +231,7 @@ static const VkPhysicalDeviceVulkan13Features recommended_vk13 = {
     .computeFullSubgroups = true,
     .maintenance4 = true,
     .shaderZeroInitializeWorkgroupMemory = true,
+    .synchronization2 = true,
 };
 
 static const VkPhysicalDeviceVulkan12Features recommended_vk12 = {
@@ -278,15 +271,8 @@ const VkPhysicalDeviceFeatures2 pl_vulkan_recommended_features = {
 };
 
 // Required features
-static const VkPhysicalDeviceVulkan13Features required_vk13 = {
-    .sType = VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_VULKAN_1_3_FEATURES,
-    .dynamicRendering = true,
-    .synchronization2 = true,
-};
-
 static const VkPhysicalDeviceVulkan12Features required_vk12 = {
     .sType = VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_VULKAN_1_2_FEATURES,
-    .pNext = (void *) &required_vk13,
     .hostQueryReset = true,
     .timelineSemaphore = true,
 };
@@ -313,8 +299,6 @@ static bool check_required_features(struct vk_ctx *vk)
         }                                                                       \
     } while (0)
 
-    CHECK_FEATURE(1, 3, dynamicRendering);
-    CHECK_FEATURE(1, 3, synchronization2);
     CHECK_FEATURE(1, 2, hostQueryReset);
     CHECK_FEATURE(1, 2, timelineSemaphore);
 
@@ -335,6 +319,7 @@ static const struct vk_fun vk_dev_funs[] = {
     PL_VK_DEV_FUN(BindBufferMemory),
     PL_VK_DEV_FUN(BindImageMemory),
     PL_VK_DEV_FUN(CmdBeginDebugUtilsLabelEXT),
+    PL_VK_DEV_FUN(CmdBeginRenderPass),
     PL_VK_DEV_FUN(CmdBindDescriptorSets),
     PL_VK_DEV_FUN(CmdBindIndexBuffer),
     PL_VK_DEV_FUN(CmdBindPipeline),
@@ -349,12 +334,14 @@ static const struct vk_fun vk_dev_funs[] = {
     PL_VK_DEV_FUN(CmdDraw),
     PL_VK_DEV_FUN(CmdDrawIndexed),
     PL_VK_DEV_FUN(CmdEndDebugUtilsLabelEXT),
+    PL_VK_DEV_FUN(CmdEndRenderPass),
+    PL_VK_DEV_FUN(CmdPipelineBarrier),
     PL_VK_DEV_FUN(CmdPushConstants),
     PL_VK_DEV_FUN(CmdResetQueryPool),
     PL_VK_DEV_FUN(CmdSetScissor),
     PL_VK_DEV_FUN(CmdSetViewport),
     PL_VK_DEV_FUN(CmdUpdateBuffer),
-    PL_VK_DEV_FUN(CmdWriteTimestamp2KHR),
+    PL_VK_DEV_FUN(CmdWriteTimestamp),
     PL_VK_DEV_FUN(CreateBuffer),
     PL_VK_DEV_FUN(CreateBufferView),
     PL_VK_DEV_FUN(CreateCommandPool),
@@ -362,12 +349,14 @@ static const struct vk_fun vk_dev_funs[] = {
     PL_VK_DEV_FUN(CreateDescriptorPool),
     PL_VK_DEV_FUN(CreateDescriptorSetLayout),
     PL_VK_DEV_FUN(CreateFence),
+    PL_VK_DEV_FUN(CreateFramebuffer),
     PL_VK_DEV_FUN(CreateGraphicsPipelines),
     PL_VK_DEV_FUN(CreateImage),
     PL_VK_DEV_FUN(CreateImageView),
     PL_VK_DEV_FUN(CreatePipelineCache),
     PL_VK_DEV_FUN(CreatePipelineLayout),
     PL_VK_DEV_FUN(CreateQueryPool),
+    PL_VK_DEV_FUN(CreateRenderPass),
     PL_VK_DEV_FUN(CreateSampler),
     PL_VK_DEV_FUN(CreateSemaphore),
     PL_VK_DEV_FUN(CreateShaderModule),
@@ -378,13 +367,14 @@ static const struct vk_fun vk_dev_funs[] = {
     PL_VK_DEV_FUN(DestroyDescriptorSetLayout),
     PL_VK_DEV_FUN(DestroyDevice),
     PL_VK_DEV_FUN(DestroyFence),
+    PL_VK_DEV_FUN(DestroyFramebuffer),
     PL_VK_DEV_FUN(DestroyImage),
     PL_VK_DEV_FUN(DestroyImageView),
-    PL_VK_DEV_FUN(DestroyInstance),
     PL_VK_DEV_FUN(DestroyPipeline),
     PL_VK_DEV_FUN(DestroyPipelineCache),
     PL_VK_DEV_FUN(DestroyPipelineLayout),
     PL_VK_DEV_FUN(DestroyQueryPool),
+    PL_VK_DEV_FUN(DestroyRenderPass),
     PL_VK_DEV_FUN(DestroySampler),
     PL_VK_DEV_FUN(DestroySemaphore),
     PL_VK_DEV_FUN(DestroyShaderModule),
@@ -401,6 +391,7 @@ static const struct vk_fun vk_dev_funs[] = {
     PL_VK_DEV_FUN(GetQueryPoolResults),
     PL_VK_DEV_FUN(InvalidateMappedMemoryRanges),
     PL_VK_DEV_FUN(MapMemory),
+    PL_VK_DEV_FUN(QueueSubmit),
     PL_VK_DEV_FUN(QueueWaitIdle),
     PL_VK_DEV_FUN(ResetFences),
     PL_VK_DEV_FUN(ResetQueryPool),
@@ -466,9 +457,6 @@ void pl_vk_inst_destroy(pl_vk_inst *inst_ptr)
     pl_free_ptr((void **) inst_ptr);
 }
 
-#ifdef MSAN
-__attribute__((no_sanitize("memory")))
-#endif
 static VkBool32 VKAPI_PTR vk_dbg_utils_cb(VkDebugUtilsMessageSeverityFlagBitsEXT sev,
                                           VkDebugUtilsMessageTypeFlagsEXT msgType,
                                           const VkDebugUtilsMessengerCallbackDataEXT *data,
@@ -476,10 +464,6 @@ static VkBool32 VKAPI_PTR vk_dbg_utils_cb(VkDebugUtilsMessageSeverityFlagBitsEXT
 {
     pl_log log = priv;
 
-    // MSAN really doesn't like reading from the stack-allocated memory
-    // allocated by the non-instrumented vulkan library, so just comment it out
-    // when building with MSAN as a cheap hack-around.
-
     // Ignore errors for messages that we consider false positives
     switch (data->messageIdNumber) {
     case 0x7cd0911d: // VUID-VkSwapchainCreateInfoKHR-imageExtent-01274
@@ -487,6 +471,7 @@ static VkBool32 VKAPI_PTR vk_dbg_utils_cb(VkDebugUtilsMessageSeverityFlagBitsEXT
     case 0xdc18ad6b: // UNASSIGNED-BestPractices-vkAllocateMemory-small-allocation
     case 0xb3d4346b: // UNASSIGNED-BestPractices-vkBindMemory-small-dedicated-allocation
     case 0x6cfe18a5: // UNASSIGNED-BestPractices-SemaphoreCount
+    case 0x48a09f6c: // UNASSIGNED-BestPractices-pipeline-stage-flags
     // profile chain expectations
     case 0x30f4ac70: // VUID-VkImageCreateInfo-pNext-06811
         return false;
@@ -565,6 +550,7 @@ pl_vk_inst pl_vk_inst_create(pl_log log, const struct pl_vk_inst_params *params)
     void *tmp = pl_tmp(NULL);
     params = PL_DEF(params, &pl_vk_inst_default_params);
     VkInstance inst = NULL;
+    pl_clock_t start;
 
     PL_ARRAY(const char *) exts = {0};
 
@@ -602,11 +588,13 @@ pl_vk_inst pl_vk_inst_create(pl_log log, const struct pl_vk_inst_params *params)
     };
 
     // Enumerate all supported layers
+    start = pl_clock_now();
     PL_VK_LOAD_FUN(NULL, EnumerateInstanceLayerProperties, get_addr);
     uint32_t num_layers_avail = 0;
     EnumerateInstanceLayerProperties(&num_layers_avail, NULL);
     VkLayerProperties *layers_avail = pl_calloc_ptr(tmp, num_layers_avail, layers_avail);
     EnumerateInstanceLayerProperties(&num_layers_avail, layers_avail);
+    pl_log_cpu_time(log, start, pl_clock_now(), "enumerating instance layers");
 
     pl_debug(log, "Available layers:");
     for (int i = 0; i < num_layers_avail; i++) {
@@ -626,6 +614,7 @@ pl_vk_inst pl_vk_inst_create(pl_log log, const struct pl_vk_inst_params *params)
     // happens (random segfaults, yum)
     bool debug = params->debug;
     uint32_t debug_layer = 0; // layer idx of debug layer
+    uint32_t debug_layer_version = 0;
     if (debug) {
         for (int i = 0; i < PL_ARRAY_SIZE(debug_layers); i++) {
             for (int n = 0; n < num_layers_avail; n++) {
@@ -633,7 +622,9 @@ pl_vk_inst pl_vk_inst_create(pl_log log, const struct pl_vk_inst_params *params)
                     continue;
 
                 debug_layer = n;
-                pl_info(log, "Enabling debug meta layer: %s", debug_layers[i]);
+                debug_layer_version = layers_avail[n].specVersion;
+                pl_info(log, "Enabling debug meta layer: %s (v%d.%d.%d)",
+                        debug_layers[i], PRINTF_VER(debug_layer_version));
                 PL_ARRAY_APPEND(tmp, layers, debug_layers[i]);
                 goto debug_layers_done;
             }
@@ -659,20 +650,8 @@ debug_layers_done: ;
         }
     }
 
-    static const char *extra_layers[] = {
-        "VK_LAYER_KHRONOS_synchronization2",
-    };
-
-    for (int i = 0; i < PL_ARRAY_SIZE(extra_layers); i++) {
-        for (int n = 0; n < num_layers_avail; n++) {
-            if (strcmp(extra_layers[i], layers_avail[n].layerName) == 0) {
-                PL_ARRAY_APPEND(tmp, layers, extra_layers[i]);
-                break;
-            }
-        }
-    }
-
     // Enumerate all supported extensions
+    start = pl_clock_now();
     PL_VK_LOAD_FUN(NULL, EnumerateInstanceExtensionProperties, get_addr);
     uint32_t num_exts_avail = 0;
     EnumerateInstanceExtensionProperties(NULL, &num_exts_avail, NULL);
@@ -702,6 +681,7 @@ debug_layers_done: ;
         }
     }
 
+    pl_log_cpu_time(log, start, pl_clock_now(), "enumerating instance extensions");
     pl_debug(log, "Available instance extensions:");
     for (int i = 0; i < num_exts_avail; i++)
         pl_debug(log, "    %s", exts_avail[i].extensionName);
@@ -812,13 +792,18 @@ next_opt_user_ext: ;
 
 debug_ext_done: ;
 
-    if (debug && params->debug_extra) {
-
+    // Limit this to 1.3.250+ because of bugs in older versions.
+    if (debug && params->debug_extra &&
+        debug_layer_version >= VK_MAKE_API_VERSION(0, 1, 3, 259))
+    {
         // Try enabling as many validation features as possible
         static const VkValidationFeatureEnableEXT validation_features[] = {
             VK_VALIDATION_FEATURE_ENABLE_GPU_ASSISTED_EXT,
             VK_VALIDATION_FEATURE_ENABLE_GPU_ASSISTED_RESERVE_BINDING_SLOT_EXT,
             VK_VALIDATION_FEATURE_ENABLE_BEST_PRACTICES_EXT,
+            // Depends on timeline semaphores being implemented:
+            // See https://github.com/KhronosGroup/Vulkan-ValidationLayers/issues/7600
+            //VK_VALIDATION_FEATURE_ENABLE_SYNCHRONIZATION_VALIDATION_EXT,
         };
 
         static const VkValidationFeaturesEXT vinfo = {
@@ -865,8 +850,10 @@ debug_extra_ext_done: ;
             pl_info(log, "    %s", layers.elem[i]);
     }
 
+    start = pl_clock_now();
     PL_VK_LOAD_FUN(NULL, CreateInstance, get_addr);
     VkResult res = CreateInstance(&info, PL_VK_ALLOC, &inst);
+    pl_log_cpu_time(log, start, pl_clock_now(), "creating vulkan instance");
     if (res != VK_SUCCESS) {
         pl_fatal(log, "Failed creating instance: %s", vk_res_str(res));
         goto error;
@@ -934,7 +921,7 @@ void pl_vulkan_destroy(pl_vulkan *pl_vk)
         }
         vk_malloc_destroy(&vk->ma);
         for (int i = 0; i < vk->pools.num; i++)
-            vk_cmdpool_destroy(vk, vk->pools.elem[i]);
+            vk_cmdpool_destroy(vk->pools.elem[i]);
 
         if (!vk->imported)
             vk->DestroyDevice(vk->dev, PL_VK_ALLOC);
@@ -992,11 +979,13 @@ VkPhysicalDevice pl_vulkan_choose_device(pl_log log,
     PL_VK_LOAD_FUN(inst, GetPhysicalDeviceProperties2, get_addr);
     pl_assert(GetPhysicalDeviceProperties2);
 
+    pl_clock_t start = pl_clock_now();
     VkPhysicalDevice *devices = NULL;
     uint32_t num = 0;
     VK(EnumeratePhysicalDevices(inst, &num, NULL));
     devices = pl_calloc_ptr(NULL, num, devices);
     VK(EnumeratePhysicalDevices(inst, &num, devices));
+    pl_log_cpu_time(log, start, pl_clock_now(), "enumerating physical devices");
 
     static const struct { const char *name; int priority; } types[] = {
         [VK_PHYSICAL_DEVICE_TYPE_DISCRETE_GPU]   = {"discrete",   5},
@@ -1187,10 +1176,12 @@ static bool device_init(struct vk_ctx *vk, const struct pl_vulkan_params *params
     }
 
     // Enumerate all supported extensions
+    pl_clock_t start = pl_clock_now();
     uint32_t num_exts_avail = 0;
     VK(vk->EnumerateDeviceExtensionProperties(vk->physd, NULL, &num_exts_avail, NULL));
     VkExtensionProperties *exts_avail = pl_calloc_ptr(tmp, num_exts_avail, exts_avail);
     VK(vk->EnumerateDeviceExtensionProperties(vk->physd, NULL, &num_exts_avail, exts_avail));
+    pl_log_cpu_time(vk->log, start, pl_clock_now(), "enumerating device extensions");
 
     PL_DEBUG(vk, "Available device extensions:");
     for (int i = 0; i < num_exts_avail; i++)
@@ -1305,7 +1296,9 @@ static bool device_init(struct vk_ctx *vk, const struct pl_vulkan_params *params
     for (int i = 0; i < vk->exts.num; i++)
         PL_INFO(vk, "    %s", vk->exts.elem[i]);
 
+    start = pl_clock_now();
     VK(vk->CreateDevice(vk->physd, &dinfo, PL_VK_ALLOC, &vk->dev));
+    pl_log_cpu_time(vk->log, start, pl_clock_now(), "creating vulkan device");
 
     // Load all mandatory device-level functions
     for (int i = 0; i < PL_ARRAY_SIZE(vk_dev_funs); i++)
