@@ -48,6 +48,8 @@ static void timer_destroy_cb(pl_gpu gpu, pl_timer timer)
     pl_free(timer);
 }
 
+VK_CB_FUNC_DEF(timer_destroy_cb);
+
 static pl_timer vk_timer_create(pl_gpu gpu)
 {
     struct pl_vk *p = PL_PRIV(gpu);
@@ -72,7 +74,7 @@ error:
 
 static void vk_timer_destroy(pl_gpu gpu, pl_timer timer)
 {
-    vk_gpu_idle_callback(gpu, (vk_cb) timer_destroy_cb, gpu, timer);
+    vk_gpu_idle_callback(gpu, VK_CB_FUNC(timer_destroy_cb), gpu, timer);
 }
 
 static uint64_t vk_timer_query(pl_gpu gpu, pl_timer timer)
@@ -133,8 +135,8 @@ static void timer_begin(pl_gpu gpu, struct vk_cmd *cmd, pl_timer timer)
         vk->ResetQueryPool(vk->dev, timer->qpool, timer->index_write, 2);
     }
 
-    vk->CmdWriteTimestamp2KHR(cmd->buf, VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT,
-                              timer->qpool, timer->index_write);
+    vk->CmdWriteTimestamp(cmd->buf, VK_PIPELINE_STAGE_ALL_COMMANDS_BIT,
+                          timer->qpool, timer->index_write);
 
     p->cmd_timer = timer;
 }
@@ -162,8 +164,8 @@ struct vk_cmd *_begin_cmd(pl_gpu gpu, enum queue_type type, const char *label,
     }
 
     if (!p->cmd || p->cmd->pool != pool) {
-        vk_cmd_submit(vk, &p->cmd);
-        p->cmd = vk_cmd_begin(vk, pool, label);
+        vk_cmd_submit(&p->cmd);
+        p->cmd = vk_cmd_begin(pool, label);
         if (!p->cmd) {
             pl_mutex_unlock(&p->recording);
             return NULL;
@@ -196,7 +198,7 @@ bool _end_cmd(pl_gpu gpu, struct vk_cmd **pcmd, bool submit)
     if (!pcmd) {
         if (submit) {
             pl_mutex_lock(&p->recording);
-            ret = vk_cmd_submit(p->vk, &p->cmd);
+            ret = vk_cmd_submit(&p->cmd);
             pl_mutex_unlock(&p->recording);
         }
         return ret;
@@ -207,11 +209,11 @@ bool _end_cmd(pl_gpu gpu, struct vk_cmd **pcmd, bool submit)
 
     if (p->cmd_timer) {
         pl_timer timer = p->cmd_timer;
-        vk->CmdWriteTimestamp2KHR(cmd->buf, VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT,
-                                  timer->qpool, timer->index_write + 1);
+        vk->CmdWriteTimestamp(cmd->buf, VK_PIPELINE_STAGE_ALL_COMMANDS_BIT,
+                              timer->qpool, timer->index_write + 1);
 
         timer->pending |= timer_bit(timer->index_write);
-        vk_cmd_callback(cmd, (vk_cb) timer_end_cb, timer,
+        vk_cmd_callback(cmd, timer_end_cb, timer,
                         (void *) (uintptr_t) timer->index_write);
 
         timer->index_write = (timer->index_write + 2) % QUERY_POOL_SIZE;
@@ -227,7 +229,7 @@ bool _end_cmd(pl_gpu gpu, struct vk_cmd **pcmd, bool submit)
         vk->CmdEndDebugUtilsLabelEXT(cmd->buf);
 
     if (submit)
-        ret = vk_cmd_submit(vk, &p->cmd);
+        ret = vk_cmd_submit(&p->cmd);
 
     pl_mutex_unlock(&p->recording);
     return ret;
@@ -252,7 +254,7 @@ static void vk_gpu_destroy(pl_gpu gpu)
     struct pl_vk *p = PL_PRIV(gpu);
     struct vk_ctx *vk = p->vk;
 
-    vk_cmd_submit(vk, &p->cmd);
+    vk_cmd_submit(&p->cmd);
     vk_wait_idle(vk);
 
     for (enum pl_tex_sample_mode s = 0; s < PL_TEX_SAMPLE_MODE_COUNT; s++) {
@@ -260,7 +262,7 @@ static void vk_gpu_destroy(pl_gpu gpu)
             vk->DestroySampler(vk->dev, p->samplers[s][a], PL_VK_ALLOC);
     }
 
-    spirv_compiler_destroy(&p->spirv);
+    pl_spirv_destroy(&p->spirv);
     pl_mutex_destroy(&p->recording);
     pl_free((void *) gpu);
 }
@@ -415,11 +417,9 @@ pl_gpu pl_gpu_create_vk(struct vk_ctx *vk)
 
     struct pl_vk *p = PL_PRIV(gpu);
     pl_mutex_init(&p->recording);
-    p->impl = pl_fns_vk;
     p->vk = vk;
-
-    const struct pl_spirv_version spirv_ver = get_spirv_version(vk);
-    p->spirv = spirv_compiler_create(vk->log, &spirv_ver);
+    p->impl = pl_fns_vk;
+    p->spirv = pl_spirv_create(vk->log, get_spirv_version(vk));
     if (!p->spirv)
         goto error;
 
@@ -465,6 +465,7 @@ pl_gpu pl_gpu_create_vk(struct vk_ctx *vk)
         if (!strcmp(vk->exts.elem[i], VK_KHR_PORTABILITY_SUBSET_EXTENSION_NAME)) {
             vk_link_struct(&props, &port_props);
             is_portability = true;
+            break;
         }
     }
 #endif
@@ -505,17 +506,20 @@ pl_gpu pl_gpu_create_vk(struct vk_ctx *vk)
         gpu->glsl.max_gather_offset = limits.maxTexelGatherOffset;
     }
 
+    const size_t max_size = vk_malloc_avail(vk->ma, 0);
     gpu->limits = (struct pl_gpu_limits) {
         // pl_gpu
         .thread_safe        = true,
         .callbacks          = true,
         // pl_buf
-        .max_buf_size       = vk_malloc_avail(vk->ma, 0),
-        .max_ubo_size       = limits.maxUniformBufferRange,
-        .max_ssbo_size      = limits.maxStorageBufferRange,
+        .max_buf_size       = max_size,
+        .max_ubo_size       = PL_MIN(limits.maxUniformBufferRange, max_size),
+        .max_ssbo_size      = PL_MIN(limits.maxStorageBufferRange, max_size),
         .max_vbo_size       = vk_malloc_avail(vk->ma, VK_MEMORY_PROPERTY_DEVICE_LOCAL_BIT),
         .max_mapped_size    = vk_malloc_avail(vk->ma, VK_MEMORY_PROPERTY_HOST_VISIBLE_BIT),
-        .max_buffer_texels  = limits.maxTexelBufferElements,
+        .max_mapped_vram    = vk_malloc_avail(vk->ma, VK_MEMORY_PROPERTY_HOST_VISIBLE_BIT |
+                                                      VK_MEMORY_PROPERTY_DEVICE_LOCAL_BIT),
+        .max_buffer_texels  = PL_MIN(limits.maxTexelBufferElements, max_size),
         .align_host_ptr     = host_props.minImportedHostPointerAlignment,
         .host_cached        = vk_malloc_avail(vk->ma, VK_MEMORY_PROPERTY_HOST_CACHED_BIT),
         // pl_tex
@@ -607,133 +611,6 @@ error:
     return NULL;
 }
 
-static void vk_sync_destroy(pl_gpu gpu, pl_sync sync)
-{
-    if (!sync)
-        return;
-
-    struct pl_vk *p = PL_PRIV(gpu);
-    struct vk_ctx *vk = p->vk;
-    struct pl_sync_vk *sync_vk = PL_PRIV(sync);
-
-#ifdef PL_HAVE_UNIX
-    if (sync->handle_type == PL_HANDLE_FD) {
-        if (sync->wait_handle.fd > -1)
-            close(sync->wait_handle.fd);
-        if (sync->signal_handle.fd > -1)
-            close(sync->signal_handle.fd);
-    }
-#endif
-#ifdef PL_HAVE_WIN32
-    if (sync->handle_type == PL_HANDLE_WIN32) {
-        if (sync->wait_handle.handle != NULL)
-            CloseHandle(sync->wait_handle.handle);
-        if (sync->signal_handle.handle != NULL)
-            CloseHandle(sync->signal_handle.handle);
-    }
-    // PL_HANDLE_WIN32_KMT is just an identifier. It doesn't get closed.
-#endif
-
-    vk->DestroySemaphore(vk->dev, sync_vk->wait, PL_VK_ALLOC);
-    vk->DestroySemaphore(vk->dev, sync_vk->signal, PL_VK_ALLOC);
-
-    pl_free((void *) sync);
-}
-
-void vk_sync_deref(pl_gpu gpu, pl_sync sync)
-{
-    if (!sync)
-        return;
-
-    struct pl_sync_vk *sync_vk = PL_PRIV(sync);
-    if (pl_rc_deref(&sync_vk->rc))
-        vk_sync_destroy(gpu, sync);
-}
-
-static pl_sync vk_sync_create(pl_gpu gpu, enum pl_handle_type handle_type)
-{
-    struct pl_vk *p = PL_PRIV(gpu);
-    struct vk_ctx *vk = p->vk;
-
-    struct pl_sync_t *sync = pl_zalloc_obj(NULL, sync, struct pl_sync_vk);
-    sync->handle_type = handle_type;
-
-    struct pl_sync_vk *sync_vk = PL_PRIV(sync);
-    pl_rc_init(&sync_vk->rc);
-
-    VkExportSemaphoreCreateInfoKHR einfo = {
-        .sType = VK_STRUCTURE_TYPE_EXPORT_SEMAPHORE_CREATE_INFO_KHR,
-        .handleTypes = vk_sync_handle_type(handle_type),
-    };
-
-    switch (handle_type) {
-    case PL_HANDLE_FD:
-        sync->wait_handle.fd = -1;
-        sync->signal_handle.fd = -1;
-        break;
-    case PL_HANDLE_WIN32:
-    case PL_HANDLE_WIN32_KMT:
-        sync->wait_handle.handle = NULL;
-        sync->signal_handle.handle = NULL;
-        break;
-    case PL_HANDLE_DMA_BUF:
-    case PL_HANDLE_HOST_PTR:
-    case PL_HANDLE_MTL_TEX:
-    case PL_HANDLE_IOSURFACE:
-        pl_unreachable();
-    }
-
-    const VkSemaphoreCreateInfo sinfo = {
-        .sType = VK_STRUCTURE_TYPE_SEMAPHORE_CREATE_INFO,
-        .pNext = &einfo,
-    };
-
-    VK(vk->CreateSemaphore(vk->dev, &sinfo, PL_VK_ALLOC, &sync_vk->wait));
-    VK(vk->CreateSemaphore(vk->dev, &sinfo, PL_VK_ALLOC, &sync_vk->signal));
-    PL_VK_NAME(SEMAPHORE, sync_vk->wait, "sync wait");
-    PL_VK_NAME(SEMAPHORE, sync_vk->signal, "sync signal");
-
-#ifdef PL_HAVE_UNIX
-    if (handle_type == PL_HANDLE_FD) {
-        VkSemaphoreGetFdInfoKHR finfo = {
-            .sType = VK_STRUCTURE_TYPE_SEMAPHORE_GET_FD_INFO_KHR,
-            .semaphore = sync_vk->wait,
-            .handleType = einfo.handleTypes,
-        };
-
-        VK(vk->GetSemaphoreFdKHR(vk->dev, &finfo, &sync->wait_handle.fd));
-
-        finfo.semaphore = sync_vk->signal;
-        VK(vk->GetSemaphoreFdKHR(vk->dev, &finfo, &sync->signal_handle.fd));
-    }
-#endif
-
-#ifdef PL_HAVE_WIN32
-    if (handle_type == PL_HANDLE_WIN32 ||
-        handle_type == PL_HANDLE_WIN32_KMT)
-    {
-        VkSemaphoreGetWin32HandleInfoKHR handle_info = {
-            .sType = VK_STRUCTURE_TYPE_SEMAPHORE_GET_WIN32_HANDLE_INFO_KHR,
-            .semaphore = sync_vk->wait,
-            .handleType = einfo.handleTypes,
-        };
-
-        VK(vk->GetSemaphoreWin32HandleKHR(vk->dev, &handle_info,
-                                          &sync->wait_handle.handle));
-
-        handle_info.semaphore = sync_vk->signal;
-        VK(vk->GetSemaphoreWin32HandleKHR(vk->dev, &handle_info,
-                                          &sync->signal_handle.handle));
-    }
-#endif
-
-    return sync;
-
-error:
-    vk_sync_destroy(gpu, sync);
-    return NULL;
-}
-
 void pl_vulkan_sem_destroy(pl_gpu gpu, VkSemaphore *semaphore)
 {
     VkSemaphore sem = *semaphore;
@@ -877,8 +754,8 @@ struct vk_cmd *pl_vk_steal_cmd(pl_gpu gpu)
 
     struct vk_cmdpool *pool = vk->pool_graphics;
     if (!cmd || cmd->pool != pool) {
-        vk_cmd_submit(vk, &cmd);
-        cmd = vk_cmd_begin(vk, pool, NULL);
+        vk_cmd_submit(&cmd);
+        cmd = vk_cmd_begin(pool, NULL);
     }
 
     return cmd;
@@ -901,7 +778,6 @@ static const struct pl_gpu_fns pl_fns_vk = {
     .tex_upload             = vk_tex_upload,
     .tex_download           = vk_tex_download,
     .tex_poll               = vk_tex_poll,
-    .tex_export             = vk_tex_export,
     .buf_create             = vk_buf_create,
     .buf_destroy            = vk_buf_deref,
     .buf_write              = vk_buf_write,
@@ -913,8 +789,6 @@ static const struct pl_gpu_fns pl_fns_vk = {
     .pass_create            = vk_pass_create,
     .pass_destroy           = vk_pass_destroy,
     .pass_run               = vk_pass_run,
-    .sync_create            = vk_sync_create,
-    .sync_destroy           = vk_sync_deref,
     .timer_create           = vk_timer_create,
     .timer_destroy          = vk_timer_destroy,
     .timer_query            = vk_timer_query,
